import argparse
import logging

import numpy as np
import torch

import openpifpaf

try:
    import pycocotools.coco
    # monkey patch for Python 3 compat
    pycocotools.coco.unicode = str
except ImportError:
    pass

LOG = logging.getLogger(__name__)


KEYPOINTS = [
    'left_shoulder',  # 1
    'right_shoulder',  # 2
    'left_elbow',  # 3
    'right_elbow',  # 4
    'left_wrist',  # 5
    'right_wrist',  # 6
    'left_hip',  # 7
    'right_hip',  # 8
    'left_knee',  # 9
    'right_knee',  # 10
    'left_ankle',  # 11
    'right_ankle',  # 12
    'head',  # 13
    'neck',  # 14
]
SKELETON = [
    [13, 14],  # head-neck
    [14, 1], [14, 2],  # neck to shoulders
    [1, 2],  # shoulders
    [7, 8],  # hips
    [1, 3], [3, 5],  # left arm
    [2, 4], [4, 6],  # right arm
    [1, 7],  # left shoulder-hip
    [2, 8],  # right shoulder-hip
    [7, 9], [9, 11],  # left leg
    [8, 10], [10, 12],  # right leg
]
# sigmas:
# https://github.com/Jeff-sjtu/CrowdPose/blob/master/
# crowdpose-api/PythonAPI/crowdposetools/cocoeval.py#L223
SIGMAS = [
    0.079,  # shoulders
    0.079,  # shoulders
    0.072,  # elbows
    0.072,  # elbows
    0.062,  # wrists
    0.062,  # wrists
    0.107,  # hips
    0.107,  # hips
    0.087,  # knees
    0.087,  # knees
    0.089,  # ankles
    0.089,  # ankles
    0.079,  # head
    0.079,  # neck
]
UPRIGHT_POSE = np.array([
    [-1.4, 8.0, 2.0],  # 'left_shoulder',
    [1.4, 8.0, 2.0],  # 'right_shoulder',
    [-1.75, 6.0, 2.0],  # 'left_elbow',
    [1.75, 6.2, 2.0],  # 'right_elbow',
    [-1.75, 4.0, 2.0],  # 'left_wrist',
    [1.75, 4.2, 2.0],  # 'right_wrist',
    [-1.26, 4.0, 2.0],  # 'left_hip',
    [1.26, 4.0, 2.0],  # 'right_hip',
    [-1.4, 2.0, 2.0],  # 'left_knee',
    [1.4, 2.1, 2.0],  # 'right_knee',
    [-1.4, 0.0, 2.0],  # 'left_ankle',
    [1.4, 0.1, 2.0],  # 'right_ankle',
    [0.0, 10.3, 2.0],  # head
    [0.0, 9.3, 2.0],  # neck,
])
HFLIP = openpifpaf.plugins.coco.constants.HFLIP
COCO_CATEGORIES = openpifpaf.plugins.coco.constants.COCO_CATEGORIES


class CrowdPose(openpifpaf.datasets.DataModule):
    _test_annotations = 'data-crowdpose/json/crowdpose_test.json'

    # cli configurable
    train_annotations = 'data-crowdpose/json/crowdpose_train.json'
    val_annotations = 'data-crowdpose/json/crowdpose_val.json'
    eval_annotations = val_annotations
    image_dir = 'data-crowdpose/images/'

    square_edge = 385
    extended_scale = False
    orientation_invariant = 0.0
    augmentation = True
    rescale_images = 1.0
    upsample_stride = 1
    min_kp_anns = 1

    eval_long_edge = 641
    eval_orientation_invariant = 0.0
    eval_extended_scale = False
    eval_crowdpose_index = None

    def __init__(self):
        super().__init__()

        cif = openpifpaf.headmeta.Cif('cif', 'crowdpose',
                                      keypoints=KEYPOINTS,
                                      sigmas=SIGMAS,
                                      pose=UPRIGHT_POSE,
                                      draw_skeleton=SKELETON)
        caf = openpifpaf.headmeta.Caf('caf', 'crowdpose',
                                      keypoints=KEYPOINTS,
                                      sigmas=SIGMAS,
                                      pose=UPRIGHT_POSE,
                                      skeleton=SKELETON)

        cif.upsample_stride = self.upsample_stride
        caf.upsample_stride = self.upsample_stride
        self.head_metas = [cif, caf]

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        group = parser.add_argument_group('data module CrowdPose')

        group.add_argument('--crowdpose-train-annotations',
                           default=cls.train_annotations)
        group.add_argument('--crowdpose-val-annotations',
                           default=cls.val_annotations)
        group.add_argument('--crowdpose-image-dir',
                           default=cls.image_dir)

        group.add_argument('--crowdpose-square-edge',
                           default=cls.square_edge, type=int,
                           help='square edge of input images')
        assert not cls.extended_scale
        group.add_argument('--crowdpose-extended-scale',
                           default=False, action='store_true',
                           help='augment with an extended scale range')
        group.add_argument('--crowdpose-orientation-invariant',
                           default=cls.orientation_invariant, type=float,
                           help='augment with random orientations')
        assert cls.augmentation
        group.add_argument('--crowdpose-no-augmentation',
                           dest='crowdpose_augmentation',
                           default=True, action='store_false',
                           help='do not apply data augmentation')
        group.add_argument('--crowdpose-rescale-images',
                           default=cls.rescale_images, type=float,
                           help='overall rescale factor for images')
        group.add_argument('--crowdpose-upsample',
                           default=cls.upsample_stride, type=int,
                           help='head upsample stride')
        group.add_argument('--crowdpose-min-kp-anns',
                           default=cls.min_kp_anns, type=int,
                           help='filter images with fewer keypoint annotations')

        # evaluation
        eval_set_group = group.add_mutually_exclusive_group()
        eval_set_group.add_argument('--crowdpose-eval-test', default=False, action='store_true')

        group.add_argument('--crowdpose-eval-long-edge', default=cls.eval_long_edge, type=int)
        assert not cls.eval_extended_scale
        group.add_argument('--crowdpose-eval-extended-scale', default=False, action='store_true')
        group.add_argument('--crowdpose-eval-orientation-invariant',
                           default=cls.eval_orientation_invariant, type=float)
        group.add_argument('--crowdpose-index', choices=('easy', 'medium', 'hard'),
                           default=None)

    @classmethod
    def configure(cls, args: argparse.Namespace):
        # extract global information
        cls.debug = args.debug
        cls.pin_memory = args.pin_memory

        # crowdpose specific
        cls.train_annotations = args.crowdpose_train_annotations
        cls.val_annotations = args.crowdpose_val_annotations
        cls.image_dir = args.crowdpose_image_dir

        cls.square_edge = args.crowdpose_square_edge
        cls.extended_scale = args.crowdpose_extended_scale
        cls.orientation_invariant = args.crowdpose_orientation_invariant
        cls.augmentation = args.crowdpose_augmentation
        cls.rescale_images = args.crowdpose_rescale_images
        cls.upsample_stride = args.crowdpose_upsample
        cls.min_kp_anns = args.crowdpose_min_kp_anns

        # evaluation
        if args.crowdpose_eval_test:
            cls.eval_annotations = cls._test_annotations
        cls.eval_long_edge = args.crowdpose_eval_long_edge
        cls.eval_orientation_invariant = args.crowdpose_eval_orientation_invariant
        cls.eval_extended_scale = args.crowdpose_eval_extended_scale
        cls.eval_crowdpose_index = args.crowdpose_index

    def _preprocess(self):
        encoders = (openpifpaf.encoder.Cif(self.head_metas[0]),
                    openpifpaf.encoder.Caf(self.head_metas[1]))

        if not self.augmentation:
            return openpifpaf.transforms.Compose([
                openpifpaf.transforms.NormalizeAnnotations(),
                openpifpaf.transforms.RescaleAbsolute(self.square_edge),
                openpifpaf.transforms.CenterPad(self.square_edge),
                openpifpaf.transforms.EVAL_TRANSFORM,
                openpifpaf.transforms.Encoders(encoders),
            ])

        if self.extended_scale:
            rescale_t = openpifpaf.transforms.RescaleRelative(
                scale_range=(0.25 * self.rescale_images,
                             2.0 * self.rescale_images),
                power_law=True, stretch_range=(0.75, 1.33))
        else:
            rescale_t = openpifpaf.transforms.RescaleRelative(
                scale_range=(0.4 * self.rescale_images,
                             2.0 * self.rescale_images),
                power_law=True, stretch_range=(0.75, 1.33))

        orientation_t = None
        if self.orientation_invariant:
            orientation_t = openpifpaf.transforms.RandomApply(
                openpifpaf.transforms.RotateBy90(), self.orientation_invariant)

        return openpifpaf.transforms.Compose([
            openpifpaf.transforms.NormalizeAnnotations(),
            openpifpaf.transforms.AnnotationJitter(),
            openpifpaf.transforms.RandomApply(openpifpaf.transforms.HFlip(KEYPOINTS, HFLIP), 0.5),
            rescale_t,
            openpifpaf.transforms.Crop(self.square_edge, use_area_of_interest=True),
            openpifpaf.transforms.CenterPad(self.square_edge),
            orientation_t,
            openpifpaf.transforms.TRAIN_TRANSFORM,
            openpifpaf.transforms.Encoders(encoders),
        ])

    def train_loader(self):
        train_data = openpifpaf.plugins.coco.CocoDataset(
            image_dir=self.image_dir,
            ann_file=self.train_annotations,
            preprocess=self._preprocess(),
            annotation_filter=True,
            min_kp_anns=self.min_kp_anns,
            category_ids=[1],
        )
        return torch.utils.data.DataLoader(
            train_data, batch_size=self.batch_size, shuffle=not self.debug,
            pin_memory=self.pin_memory, num_workers=self.loader_workers, drop_last=True,
            collate_fn=openpifpaf.datasets.collate_images_targets_meta)

    def val_loader(self):
        val_data = openpifpaf.plugins.coco.CocoDataset(
            image_dir=self.image_dir,
            ann_file=self.val_annotations,
            preprocess=self._preprocess(),
            annotation_filter=True,
            min_kp_anns=self.min_kp_anns,
            category_ids=[1],
        )
        return torch.utils.data.DataLoader(
            val_data, batch_size=self.batch_size, shuffle=False,
            pin_memory=self.pin_memory, num_workers=self.loader_workers, drop_last=True,
            collate_fn=openpifpaf.datasets.collate_images_targets_meta)

    @classmethod
    def common_eval_preprocess(cls):
        rescale_t = None
        if cls.eval_extended_scale:
            assert cls.eval_long_edge
            rescale_t = [
                openpifpaf.transforms.DeterministicEqualChoice([
                    openpifpaf.transforms.RescaleAbsolute(cls.eval_long_edge),
                    openpifpaf.transforms.RescaleAbsolute((cls.eval_long_edge - 1) // 2 + 1),
                ], salt=1)
            ]
        elif cls.eval_long_edge:
            rescale_t = openpifpaf.transforms.RescaleAbsolute(cls.eval_long_edge)

        if cls.batch_size == 1:
            padding_t = openpifpaf.transforms.CenterPadTight(16)
        else:
            assert cls.eval_long_edge
            padding_t = openpifpaf.transforms.CenterPad(cls.eval_long_edge)

        orientation_t = None
        if cls.eval_orientation_invariant:
            orientation_t = openpifpaf.transforms.DeterministicEqualChoice([
                None,
                openpifpaf.transforms.RotateBy90(fixed_angle=90),
                openpifpaf.transforms.RotateBy90(fixed_angle=180),
                openpifpaf.transforms.RotateBy90(fixed_angle=270),
            ], salt=3)

        return [
            openpifpaf.transforms.NormalizeAnnotations(),
            rescale_t,
            padding_t,
            orientation_t,
        ]

    def _eval_preprocess(self):
        return openpifpaf.transforms.Compose([
            *self.common_eval_preprocess(),
            openpifpaf.transforms.ToAnnotations([
                openpifpaf.transforms.ToKpAnnotations(
                    COCO_CATEGORIES,
                    keypoints_by_category={1: self.head_metas[0].keypoints},
                    skeleton_by_category={1: self.head_metas[1].skeleton},
                ),
                openpifpaf.transforms.ToCrowdAnnotations(COCO_CATEGORIES),
            ]),
            openpifpaf.transforms.EVAL_TRANSFORM,
        ])

    @staticmethod
    def _filter_crowdindex(data: openpifpaf.plugins.coco.CocoDataset, min_index, max_index):
        filtered_ids = []
        for id_ in data.ids:
            image_info = data.coco.imgs[id_]
            LOG.debug('image info %s', image_info)
            crowdindex = image_info['crowdIndex']
            if min_index <= crowdindex < max_index:
                filtered_ids.append(id_)

        LOG.info('crowdindex filter from %d to %d images', len(data.ids), len(filtered_ids))
        data.ids = filtered_ids

    def eval_loader(self):
        eval_data = openpifpaf.plugins.coco.CocoDataset(
            image_dir=self.image_dir,
            ann_file=self.eval_annotations,
            preprocess=self._eval_preprocess(),
            annotation_filter=(self.eval_annotations == self.val_annotations),
            min_kp_anns=self.min_kp_anns if self.eval_annotations == self.val_annotations else 0,
            category_ids=[1],
        )
        if self.eval_crowdpose_index == 'easy':
            self._filter_crowdindex(eval_data, 0.0, 0.1)
        elif self.eval_crowdpose_index == 'medium':
            self._filter_crowdindex(eval_data, 0.1, 0.8)
        elif self.eval_crowdpose_index == 'hard':
            self._filter_crowdindex(eval_data, 0.8, 1.0)

        return torch.utils.data.DataLoader(
            eval_data, batch_size=self.batch_size, shuffle=False,
            pin_memory=self.pin_memory, num_workers=self.loader_workers, drop_last=False,
            collate_fn=openpifpaf.datasets.collate_images_anns_meta)

    def metrics(self):
        return [openpifpaf.metric.Coco(
            pycocotools.coco.COCO(self.eval_annotations),
            max_per_image=20,
            category_ids=[1],
            iou_type='keypoints',
            keypoint_oks_sigmas=SIGMAS,
        )]



from contextlib import contextmanager
import logging
from typing import List

import numpy as np

from .. import annotation, show

try:
    import matplotlib.pyplot as plt
    from mpl_toolkits.axes_grid1 import make_axes_locatable
except ImportError:
    plt = None
    make_axes_locatable = None

LOG = logging.getLogger(__name__)
IMAGENOTGIVEN = object()


def itemsetter(list_, index, value):
    list_[index] = value
    return list_


class Base:
    all_indices = []
    common_ax = None
    processed_image_intensity_spread = 2.0

    _image = None
    _processed_image = None
    _image_meta = None
    _ground_truth: List[annotation.Base] = None

    def __init__(self, head_name):
        self.head_name = head_name
        self._ax = None

        LOG.debug('%s: indices = %s', head_name, self.indices())

    @classmethod
    def image(cls, image=IMAGENOTGIVEN, meta=None):
        if image is IMAGENOTGIVEN:  # getter
            if callable(Base._image):  # check whether stored value is lazy
                Base._image = Base._image()  # pylint: disable=not-callable
            return Base._image

        if image is None:
            Base._image = None
            Base._image_meta = None
            return cls

        Base._image = lambda: np.asarray(image)
        Base._image_meta = meta
        return cls

    @classmethod
    def processed_image(cls, image=IMAGENOTGIVEN):
        if image is IMAGENOTGIVEN:  # getter
            if callable(Base._processed_image):  # check whether stored value is lazy
                Base._processed_image = Base._processed_image()  # pylint: disable=not-callable
            return Base._processed_image

        if image is None:
            Base._processed_image = None
            return cls

        def process_image(image):
            image = np.moveaxis(np.asarray(image), 0, -1)
            image = np.clip(image / cls.processed_image_intensity_spread * 0.5 + 0.5, 0.0, 1.0)
            return image

        Base._processed_image = lambda: process_image(image)
        return cls

    @staticmethod
    def ground_truth(ground_truth):
        Base._ground_truth = ground_truth

    @staticmethod
    def reset():
        Base._image = None
        Base._image_meta = None
        Base._processed_image = None
        Base._ground_truth = None

    @classmethod
    def normalized_index(cls, data):
        if isinstance(data, str):
            data = data.split(':')

        # unpack comma separation
        for di, d in enumerate(data):
            if ',' not in d:
                continue
            multiple = [cls.normalized_index(itemsetter(data, di, unpacked))
                        for unpacked in d.split(',')]
            # flatten before return
            return [item for items in multiple for item in items]

        if len(data) >= 2 and len(data[1]) == 0:
            data[1] = -1

        if len(data) == 3:
            return [(data[0], int(data[1]), data[2])]
        if len(data) == 2:
            return [(data[0], int(data[1]), 'all')]
        return [(data[0], -1, 'all')]

    @classmethod
    def set_all_indices(cls, all_indices):
        cls.all_indices = [d for dd in all_indices for d in cls.normalized_index(dd)]

    def indices(self, type_=None, with_all=True):
        head_names = self.head_name
        if not isinstance(head_names, (tuple, list)):
            head_names = (head_names,)
        return [
            f for hn, f, r_type in self.all_indices
            if hn in head_names and (
                type_ is None
                or (with_all and r_type == 'all')
                or type_ == r_type
            )
        ]

    @staticmethod
    def colorbar(ax, colored_element, size='3%', pad=0.01):
        divider = make_axes_locatable(ax)
        cax = divider.append_axes('right', size=size, pad=pad)
        cb = plt.colorbar(colored_element, cax=cax)
        cb.outline.set_linewidth(0.1)

    @contextmanager
    def image_canvas(self, image, *args, **kwargs):
        ax = self._ax or self.common_ax
        if ax is not None:
            ax.set_axis_off()
            ax.imshow(np.asarray(image))
            yield ax
            return

        with show.image_canvas(image, *args, **kwargs) as ax:
            yield ax

    @contextmanager
    def canvas(self, *args, **kwargs):
        ax = self._ax or self.common_ax
        if ax is not None:
            yield ax
            return

        with show.canvas(*args, **kwargs) as ax:
            yield ax

    @staticmethod
    def scale_scalar(field, stride):
        field = np.repeat(field, stride, 0)
        field = np.repeat(field, stride, 1)

        # center (the result is technically still off by half a pixel)
        half_stride = stride // 2
        return field[half_stride:-half_stride + 1, half_stride:-half_stride + 1]



"""Video demo application.

Use --scale=0.2 to reduce the input image size to 20%.
Use --json-output for headless processing.

Example commands:
    python3 -m pifpaf.video --source=0  # default webcam
    python3 -m pifpaf.video --source=1  # another webcam

    # streaming source
    python3 -m pifpaf.video --source=http://127.0.0.1:8080/video

    # file system source (any valid OpenCV source)
    python3 -m pifpaf.video --source=docs/coco/000000081988.jpg

Trouble shooting:
* MacOSX: try to prefix the command with "MPLBACKEND=MACOSX".
"""


import argparse
import json
import logging
import os
import time

import torch

from . import decoder, logger, network, show, visualizer, __version__
from .predictor import Predictor
from .stream import Stream

LOG = logging.getLogger(__name__)


class CustomFormatter(argparse.ArgumentDefaultsHelpFormatter,
                      argparse.RawDescriptionHelpFormatter):
    pass


def cli():  # pylint: disable=too-many-statements,too-many-branches
    parser = argparse.ArgumentParser(
        prog='python3 -m openpifpaf.video',
        usage='%(prog)s [options]',
        description=__doc__,
        formatter_class=CustomFormatter,
    )
    parser.add_argument('--version', action='version',
                        version='OpenPifPaf {version}'.format(version=__version__))

    network.Factory.cli(parser)
    decoder.cli(parser)
    logger.cli(parser)
    Predictor.cli(parser)
    show.cli(parser)
    Stream.cli(parser)
    visualizer.cli(parser)

    parser.add_argument('--source', default='0',
                        help=('OpenCV source url. Integer for webcams. '
                              'Or ipwebcam urls (rtsp/rtmp). '
                              'Use "screen" for screen grabs.'))
    parser.add_argument('--video-output', default=None, nargs='?', const=True,
                        help='video output file or "virtualcam"')
    parser.add_argument('--json-output', default=None, nargs='?', const=True,
                        help='json output file')
    parser.add_argument('--separate-debug-ax', default=False, action='store_true')
    parser.add_argument('--disable-cuda', action='store_true',
                        help='disable CUDA')
    args = parser.parse_args()

    logger.configure(args, LOG)  # logger first

    # add args.device
    args.device = torch.device('cpu')
    args.pin_memory = False
    if not args.disable_cuda and torch.cuda.is_available():
        args.device = torch.device('cuda')
        args.pin_memory = True
    LOG.info('neural network device: %s (CUDA available: %s, count: %d)',
             args.device, torch.cuda.is_available(), torch.cuda.device_count())

    decoder.configure(args)
    network.Factory.configure(args)
    Predictor.configure(args)
    show.configure(args)
    Stream.configure(args)
    visualizer.configure(args)

    # check whether source should be an int
    if len(args.source) == 1:
        args.source = int(args.source)

    # standard filenames
    if args.video_output is True:
        args.video_output = '{}.openpifpaf.mp4'.format(args.source)
        if os.path.exists(args.video_output):
            os.remove(args.video_output)
    assert args.video_output is None or not os.path.exists(args.video_output)
    if args.json_output is True:
        args.json_output = '{}.openpifpaf.json'.format(args.source)
        if os.path.exists(args.json_output):
            os.remove(args.json_output)
    assert args.json_output is None or not os.path.exists(args.json_output)

    return args


def main():
    args = cli()

    Predictor.loader_workers = 1
    predictor = Predictor(
        visualize_image=(not args.json_output or args.video_output),
        visualize_processed_image=args.debug,
    )
    capture = Stream(args.source, preprocess=predictor.preprocess)

    annotation_painter = show.AnnotationPainter()
    animation = show.AnimationFrame(
        video_output=args.video_output,
        second_visual=args.separate_debug_ax,
    )
    if not args.json_output or args.video_output:
        ax, ax_second = animation.frame_init()
        visualizer.Base.common_ax = ax_second if args.separate_debug_ax else ax

    last_loop = time.perf_counter()
    for (ax, ax_second), (preds, _, meta) in \
            zip(animation.iter(), predictor.dataset(capture)):
        start_post = time.perf_counter()
        if args.json_output:
            with open(args.json_output, 'a+') as f:
                json.dump({
                    'frame': meta['frame_i'],
                    'predictions': [ann.json_data() for ann in preds]
                }, f, separators=(',', ':'))
                f.write('\n')
        if (not args.json_output or args.video_output) \
           and (args.separate_debug_ax or not args.debug_indices):
            image = visualizer.Base.image()
            ax.imshow(image)
            annotation_painter.annotations(ax, preds)
        postprocessing_time = time.perf_counter() - start_post
        if animation.last_draw_time is not None:
            postprocessing_time += animation.last_draw_time

        LOG.info('frame %d, loop time = %.0fms (pre = %.1fms, post = %.1fms), FPS = %.1f',
                 meta['frame_i'],
                 (time.perf_counter() - last_loop) * 1000.0,
                 meta['preprocessing_s'] * 1000.0,
                 postprocessing_time * 1000.0,
                 1.0 / (time.perf_counter() - last_loop))
        last_loop = time.perf_counter()


if __name__ == '__main__':
    main()




import openpifpaf

from . import animal_kp

from .animal_kp import AnimalKp


def register():
    openpifpaf.DATAMODULES['animal'] = animal_kp.AnimalKp
    openpifpaf.CHECKPOINT_URLS['shufflenetv2k30-animalpose'] = \
        "http://github.com/vita-epfl/openpifpaf-torchhub/releases/" \
        "download/v0.12.9/shufflenetv2k30-210511-120906-animal.pkl.epoch400"



import itertools
import logging
from typing import List

import torch

from .preprocess import Preprocess

LOG = logging.getLogger(__name__)


class RandomApply(Preprocess):
    """Randomly apply another transformation.

    :param transform: another transformation
    :param probability: probability to apply the given transform
    """
    def __init__(self, transform: Preprocess, probability: float):
        self.transform = transform
        self.probability = probability

    def __call__(self, image, anns, meta):
        if float(torch.rand(1).item()) > self.probability:
            return image, anns, meta
        return self.transform(image, anns, meta)


class RandomChoice(Preprocess):
    """Choose a single random transform."""
    def __init__(self, transforms: List[Preprocess], probabilities: List[float]):
        if sum(probabilities) < 1.0 and len(transforms) == len(probabilities):
            transforms.append(None)
        self.transforms = transforms

        if len(transforms) == len(probabilities) + 1:
            probabilities.append(1.0 - sum(probabilities))
        assert sum(probabilities) == 1.0, [transforms, probabilities]
        assert len(transforms) == len(probabilities)
        self.probabilities = probabilities

    def __call__(self, image, anns, meta):
        rnd = float(torch.rand(1).item())
        for t, p_cumulative in zip(self.transforms, itertools.accumulate(self.probabilities)):
            if rnd > p_cumulative:
                continue

            if t is None:
                return image, anns, meta
            return t(image, anns, meta)

        raise Exception('not possible')


class DeterministicEqualChoice(Preprocess):
    """Deterministically choose one of the transforms.

    :param transforms: a list of transforms
    :param salt: integer that combined with meta['image_id] determines the choice of the transform
    """
    def __init__(self, transforms: List[Preprocess], salt: int = 0):
        self.transforms = transforms
        self.salt = salt

    def __call__(self, image, anns, meta):
        assert meta['image_id'] > 0
        LOG.debug('image id = %d', meta['image_id'])
        choice = hash(meta['image_id'] + self.salt) % len(self.transforms)
        t = self.transforms[choice]
        if t is None:
            return image, anns, meta
        return t(image, anns, meta)



import logging

from .animation_frame import AnimationFrame
from .canvas import Canvas
from .painters import KeypointPainter

LOG = logging.getLogger(__name__)


def cli(parser):
    group = parser.add_argument_group('show')
    group.add_argument('--save-all', nargs='?', default=None, const='all-images/',
                       help='every plot is saved (optional to specify directory)')
    assert not Canvas.show
    assert not AnimationFrame.show
    group.add_argument('--show', default=False, action='store_true',
                       help='show every plot, i.e., call matplotlib show()')
    group.add_argument('--image-width', default=None, type=float,
                       help='image width for matplotlib (in inches)')
    group.add_argument('--image-height', default=None, type=float,
                       help='image height for matplotlib (in inches)')
    group.add_argument('--image-dpi-factor', default=Canvas.image_dpi_factor, type=float,
                       help='increase dpi of output image by this factor')
    group.add_argument('--image-min-dpi', default=Canvas.image_min_dpi, type=float,
                       help='minimum dpi of image output')
    group.add_argument('--show-file-extension', default=Canvas.out_file_extension,
                       help='default file extension')
    group.add_argument('--textbox-alpha', default=KeypointPainter.textbox_alpha, type=float,
                       help='transparency of annotation text box')
    group.add_argument('--text-color', default=KeypointPainter.text_color,
                       help='annotation text color')
    group.add_argument('--font-size', default=KeypointPainter.font_size, type=int,
                       help='annotation font size')
    assert not KeypointPainter.monocolor_connections
    group.add_argument('--monocolor-connections', default=False, action='store_true',
                       help='use a single color per instance')
    group.add_argument('--line-width', default=None, type=int,
                       help='skeleton line width')
    group.add_argument('--skeleton-solid-threshold',
                       default=KeypointPainter.solid_threshold, type=float,
                       help='set to 0.0 to draw all connections as solid lines')

    group.add_argument('--show-box', default=False, action='store_true',
                       help='show annotation bounding boxes')
    group.add_argument('--white-overlay',
                       nargs='?', default=False, const=0.8, type=float,
                       help='increase contrast to annotations by making image whiter')
    group.add_argument('--show-joint-scales', default=False, action='store_true',
                       help='show boxes representing joint sizes')
    group.add_argument('--show-joint-confidences', default=False, action='store_true',
                       help='print per-joint confidences on skeleton annotations')
    group.add_argument('--show-decoding-order', default=False, action='store_true')
    group.add_argument('--show-frontier-order', default=False, action='store_true')
    group.add_argument('--show-only-decoded-connections', default=False, action='store_true',
                       help='to debug which connections were used for decoding')

    group.add_argument('--video-fps', default=AnimationFrame.video_fps, type=float,
                       help='output video frame rate (frames per second)')
    group.add_argument('--video-dpi', default=AnimationFrame.video_dpi, type=float,
                       help='output video resolution (dots per inch)')


def configure(args):
    Canvas.all_images_directory = args.save_all
    Canvas.show = args.show
    if args.image_width is not None:
        Canvas.image_width = args.image_width
    if args.image_height is not None:
        Canvas.image_height = args.image_height
        if args.image_width is None:
            # if only image height is provided, do not force image width
            Canvas.image_width = None
    Canvas.image_dpi_factor = args.image_dpi_factor
    Canvas.white_overlay = args.white_overlay
    Canvas.image_min_dpi = args.image_min_dpi
    Canvas.out_file_extension = args.show_file_extension

    KeypointPainter.show_box = args.show_box
    KeypointPainter.show_joint_scales = args.show_joint_scales
    KeypointPainter.show_joint_confidences = args.show_joint_confidences
    KeypointPainter.show_decoding_order = args.show_decoding_order
    KeypointPainter.show_frontier_order = args.show_frontier_order
    KeypointPainter.show_only_decoded_connections = args.show_only_decoded_connections

    KeypointPainter.textbox_alpha = args.textbox_alpha
    KeypointPainter.text_color = args.text_color
    KeypointPainter.monocolor_connections = args.monocolor_connections
    KeypointPainter.line_width = args.line_width
    KeypointPainter.solid_threshold = args.skeleton_solid_threshold
    KeypointPainter.font_size = args.font_size

    AnimationFrame.video_fps = args.video_fps
    AnimationFrame.video_dpi = args.video_dpi
    AnimationFrame.show = args.show



import argparse

import PIL
import torch

import openpifpaf
from openpifpaf.transforms.pair import SingleImage as S

from . import datasets, metric
from .normalize_transform import NormalizePosetrack

from .constants import (
    KEYPOINTS,
    SIGMAS,
    UPRIGHT_POSE,
    SKELETON,
    DENSER_CONNECTIONS,
)


class Posetrack2018(openpifpaf.datasets.DataModule):
    # cli configurable
    train_annotations = 'data-posetrack2018/annotations/train/*.json'
    val_annotations = 'data-posetrack2018/annotations/val/*.json'
    eval_annotations = val_annotations
    data_root = 'data-posetrack2018'

    square_edge = 385
    with_dense = False
    augmentation = True
    rescale_images = 1.0
    upsample_stride = 1
    min_kp_anns = 1
    bmin = 0.1
    sample_pairing = 0.0
    image_aug = 0.0
    max_shift = 30.0

    eval_long_edge = 801
    eval_orientation_invariant = 0.0
    eval_extended_scale = False

    ablation_without_tcaf = False

    def __init__(self):
        super().__init__()

        cif = openpifpaf.headmeta.TSingleImageCif(
            'cif', 'posetrack2018',
            keypoints=KEYPOINTS,
            sigmas=SIGMAS,
            pose=UPRIGHT_POSE,
            draw_skeleton=SKELETON,
        )
        caf = openpifpaf.headmeta.TSingleImageCaf(
            'caf', 'posetrack2018',
            keypoints=KEYPOINTS,
            sigmas=SIGMAS,
            pose=UPRIGHT_POSE,
            skeleton=SKELETON,
        )
        dcaf = openpifpaf.headmeta.TSingleImageCaf(
            'dcaf', 'posetrack2018',
            keypoints=KEYPOINTS,
            sigmas=SIGMAS,
            pose=UPRIGHT_POSE,
            skeleton=DENSER_CONNECTIONS,
            sparse_skeleton=SKELETON,
            only_in_field_of_view=True,
        )
        tcaf = openpifpaf.headmeta.Tcaf(
            'tcaf', 'posetrack2018',
            keypoints_single_frame=KEYPOINTS,
            sigmas_single_frame=SIGMAS,
            pose_single_frame=UPRIGHT_POSE,
            draw_skeleton_single_frame=SKELETON,
            only_in_field_of_view=True,
        )

        cif.upsample_stride = self.upsample_stride
        caf.upsample_stride = self.upsample_stride
        dcaf.upsample_stride = self.upsample_stride
        tcaf.upsample_stride = self.upsample_stride
        self.head_metas = [cif, caf, dcaf, tcaf] if self.with_dense else [cif, caf, tcaf]

        if self.ablation_without_tcaf:
            self.head_metas = [cif, caf, dcaf]

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        group2018 = parser.add_argument_group('data module Posetrack2018')
        group2018.add_argument('--posetrack2018-train-annotations',
                               default=cls.train_annotations,
                               help='train annotations')
        group2018.add_argument('--posetrack2018-val-annotations',
                               default=cls.val_annotations,
                               help='val annotations')
        group2018.add_argument('--posetrack2018-eval-annotations',
                               default=cls.eval_annotations,
                               help='eval annotations')
        group2018.add_argument('--posetrack2018-data-root',
                               default=cls.data_root,
                               help='data root')

        group = parser.add_argument_group('data module Posetrack')
        group.add_argument('--posetrack-square-edge',
                           default=cls.square_edge, type=int,
                           help='square edge of input images')
        assert not cls.with_dense
        group.add_argument('--posetrack-with-dense',
                           default=False, action='store_true',
                           help='train with dense connections')
        assert cls.augmentation
        group.add_argument('--posetrack-no-augmentation',
                           dest='posetrack_augmentation',
                           default=True, action='store_false',
                           help='do not apply data augmentation')
        group.add_argument('--posetrack-rescale-images',
                           default=cls.rescale_images, type=float,
                           help='overall rescale factor for images')
        group.add_argument('--posetrack-upsample',
                           default=cls.upsample_stride, type=int,
                           help='head upsample stride')
        group.add_argument('--posetrack-min-kp-anns',
                           default=cls.min_kp_anns, type=int,
                           help='filter images with fewer keypoint annotations')
        group.add_argument('--posetrack-bmin', default=cls.bmin, type=float)
        group.add_argument('--posetrack-sample-pairing', default=cls.sample_pairing,
                           type=float,
                           help='blend two samples together')
        group.add_argument('--posetrack-image-augmentations', default=cls.image_aug,
                           type=float,
                           help='autocontrast, equalize, invert, solarize')
        group.add_argument('--posetrack-max-shift', default=cls.max_shift, type=float,
                           help='max shift')

        group.add_argument('--posetrack-eval-long-edge', default=cls.eval_long_edge, type=int)
        assert not cls.eval_extended_scale
        group.add_argument('--posetrack-eval-extended-scale', default=False, action='store_true')
        group.add_argument('--posetrack-eval-orientation-invariant',
                           default=cls.eval_orientation_invariant, type=float)

        group.add_argument('--posetrack-ablation-without-tcaf', default=False, action='store_true')

    @classmethod
    def configure(cls, args: argparse.Namespace):
        # extract global information
        cls.debug = args.debug
        cls.pin_memory = args.pin_memory

        # posetrack2018 specific
        cls.train_annotations = args.posetrack2018_train_annotations
        cls.val_annotations = args.posetrack2018_val_annotations
        cls.eval_annotations = args.posetrack2018_eval_annotations
        cls.data_root = args.posetrack2018_data_root

        # common posetrack
        cls.square_edge = args.posetrack_square_edge
        cls.with_dense = args.posetrack_with_dense
        cls.augmentation = args.posetrack_augmentation
        cls.rescale_images = args.posetrack_rescale_images
        cls.upsample_stride = args.posetrack_upsample
        cls.min_kp_anns = args.posetrack_min_kp_anns
        cls.bmin = args.posetrack_bmin
        cls.sample_pairing = args.posetrack_sample_pairing
        cls.image_aug = args.posetrack_image_augmentations
        cls.max_shift = args.posetrack_max_shift

        # evaluation
        cls.eval_long_edge = args.posetrack_eval_long_edge
        cls.eval_orientation_invariant = args.posetrack_eval_orientation_invariant
        cls.eval_extended_scale = args.posetrack_eval_extended_scale

        # ablation
        cls.ablation_without_tcaf = args.posetrack_ablation_without_tcaf

    def _preprocess(self):
        encoders = [
            openpifpaf.encoder.SingleImage(
                openpifpaf.encoder.Cif(self.head_metas[0], bmin=self.bmin)),
            openpifpaf.encoder.SingleImage(
                openpifpaf.encoder.Caf(self.head_metas[1], bmin=self.bmin)),
        ]
        if self.with_dense:
            encoders.append(
                openpifpaf.encoder.SingleImage(
                    openpifpaf.encoder.Caf(self.head_metas[2], bmin=self.bmin))
            )
        if not self.ablation_without_tcaf:
            encoders.append(openpifpaf.encoder.Tcaf(self.head_metas[-1], bmin=self.bmin))

        return openpifpaf.transforms.Compose([
            *self.common_preprocess(),
            openpifpaf.transforms.pair.Encoders(encoders),
        ])

    @classmethod
    def common_preprocess(cls):
        if not cls.augmentation:
            return [
                openpifpaf.transforms.NormalizeAnnotations(),
                openpifpaf.transforms.RescaleAbsolute(cls.square_edge),
                openpifpaf.transforms.CenterPad(cls.square_edge),
                openpifpaf.transforms.EVAL_TRANSFORM,
            ]

        hflip_posetrack = openpifpaf.transforms.HFlip(
            KEYPOINTS,
            openpifpaf.plugins.coco.constants.HFLIP)
        return [
            S(NormalizePosetrack()),
            openpifpaf.transforms.RandomApply(
                openpifpaf.transforms.pair.RandomizeOneFrame(), 0.2),
            S(openpifpaf.transforms.AddCrowdForIncompleteHead()),
            S(openpifpaf.transforms.RandomApply(hflip_posetrack, 0.5)),
            S(openpifpaf.transforms.RescaleRelative(
                (0.5, 2.0), power_law=True, absolute_reference=801, stretch_range=(0.75, 1.33))),
            S(openpifpaf.transforms.RandomChoice(
                [openpifpaf.transforms.RotateBy90(angle_perturbation=30.0, prepad=True),
                 openpifpaf.transforms.RotateUniform(30.0)],
                [0.25],
            )),
            openpifpaf.transforms.pair.Crop(cls.square_edge, max_shift=cls.max_shift),
            openpifpaf.transforms.pair.Pad(cls.square_edge, max_shift=cls.max_shift),
            openpifpaf.transforms.RandomApply(
                openpifpaf.transforms.pair.SamplePairing(), cls.sample_pairing),
            S(openpifpaf.transforms.RandomChoice([
                openpifpaf.transforms.ImageTransform(PIL.ImageOps.autocontrast),
                openpifpaf.transforms.ImageTransform(PIL.ImageOps.equalize),
                openpifpaf.transforms.ImageTransform(PIL.ImageOps.invert),
                openpifpaf.transforms.ImageTransform(PIL.ImageOps.solarize),
            ], [cls.image_aug / 4, cls.image_aug / 4, cls.image_aug / 4, cls.image_aug / 4])),
            S(openpifpaf.transforms.TRAIN_TRANSFORM),
        ]

    def train_loader(self):
        train_data = datasets.Posetrack2018(
            annotation_files=self.train_annotations,
            data_root=self.data_root,
            group=[(0, -12), (0, -8), (0, -4)],
            preprocess=self._preprocess(),
            only_annotated=True,
        )

        # to keep base-net batch size equal across batches, train tracking with
        # half the batch-size of single-image datasets
        assert self.batch_size % 2 == 0
        return torch.utils.data.DataLoader(
            train_data, batch_size=self.batch_size // 2, shuffle=not self.debug,
            pin_memory=self.pin_memory, num_workers=self.loader_workers, drop_last=True,
            collate_fn=openpifpaf.datasets.collate_tracking_images_targets_meta)

    def val_loader(self):
        val_data = datasets.Posetrack2018(
            annotation_files=self.val_annotations,
            data_root=self.data_root,
            group=[(0, -12), (0, -8), (0, -4)],
            preprocess=self._preprocess(),
            only_annotated=True,
        )

        # to keep base-net batch size equal across batches, train tracking with
        # half the batch-size of single-image datasets
        assert self.batch_size % 2 == 0
        return torch.utils.data.DataLoader(
            val_data, batch_size=self.batch_size // 2, shuffle=not self.debug,
            pin_memory=self.pin_memory, num_workers=self.loader_workers, drop_last=True,
            collate_fn=openpifpaf.datasets.collate_tracking_images_targets_meta)

    @classmethod
    def common_eval_preprocess(cls):
        rescale_t = None
        if cls.eval_extended_scale:
            assert cls.eval_long_edge
            rescale_t = [
                openpifpaf.transforms.DeterministicEqualChoice([
                    openpifpaf.transforms.RescaleAbsolute(cls.eval_long_edge),
                    openpifpaf.transforms.RescaleAbsolute((cls.eval_long_edge - 1) // 2 + 1),
                ], salt=1)
            ]
        elif cls.eval_long_edge:
            rescale_t = openpifpaf.transforms.RescaleAbsolute(cls.eval_long_edge)

        if cls.batch_size == 1:
            padding_t = openpifpaf.transforms.CenterPadTight(16)
        else:
            assert cls.eval_long_edge
            padding_t = openpifpaf.transforms.CenterPad(cls.eval_long_edge)

        orientation_t = None
        if cls.eval_orientation_invariant:
            orientation_t = openpifpaf.transforms.DeterministicEqualChoice([
                None,
                openpifpaf.transforms.RotateBy90(fixed_angle=90),
                openpifpaf.transforms.RotateBy90(fixed_angle=180),
                openpifpaf.transforms.RotateBy90(fixed_angle=270),
            ], salt=3)

        return [
            openpifpaf.transforms.pair.Ungroup(),
            NormalizePosetrack(),
            rescale_t,
            padding_t,
            orientation_t,
        ]

    def _eval_preprocess(self):
        return openpifpaf.transforms.Compose([
            *self.common_eval_preprocess(),
            openpifpaf.transforms.ToAnnotations([
                openpifpaf.transforms.ToKpAnnotations(
                    ['person'],
                    keypoints_by_category={1: self.head_metas[0].keypoints},
                    skeleton_by_category={1: self.head_metas[1].skeleton},
                ),
                openpifpaf.transforms.ToCrowdAnnotations(['person']),
            ]),
            openpifpaf.transforms.EVAL_TRANSFORM,
        ])

    def eval_loader(self):
        eval_data = datasets.Posetrack2018(
            annotation_files=self.eval_annotations,
            data_root=self.data_root,
            preprocess=self._eval_preprocess(),
        )
        eval_loader = torch.utils.data.DataLoader(
            eval_data, batch_size=self.batch_size, shuffle=False,
            pin_memory=self.pin_memory, num_workers=self.loader_workers, drop_last=False,
            collate_fn=openpifpaf.datasets.collate_images_anns_meta)
        return openpifpaf.datasets.LoaderWithReset(eval_loader, 'annotation_file')

    def metrics(self):
        eval_data = datasets.Posetrack2018(
            annotation_files=self.eval_annotations,
            data_root=self.data_root,
            preprocess=self._eval_preprocess(),
        )
        return [metric.Posetrack(
            images=eval_data.meta_images(),
            categories=eval_data.meta_categories(),
            ground_truth=self.eval_annotations,
        )]



import logging

from .base import Base
from .. import headmeta, show

LOG = logging.getLogger(__name__)


class MultiTracking(Base):
    trail_length = 10

    def __init__(self, meta: headmeta.Caf):
        super().__init__('multi_' + meta.name)
        LOG.debug('vis %s', 'multi_' + meta.name)
        self.meta = meta
        self.annotation_painter = show.AnnotationPainter()

        self.anns_trail = []

    def predicted(self, anns):
        if not self.indices():
            return

        self.anns_trail.append(anns)
        if len(self.anns_trail) > self.trail_length:
            self.anns_trail.pop(0)

        current_ids = {ann.id_ for ann in self.anns_trail[-1]}
        with self.image_canvas(self._image) as ax:
            for frame_i, frame_anns in enumerate(self.anns_trail):
                # only show trails for poses that are in the current frame
                frame_anns = [ann for ann in frame_anns if ann.id_ in current_ids]

                # only show trails for poses that have confidence > 0.01
                frame_anns = [ann for ann in frame_anns if ann.score > 0.01]

                alpha = 0.5**(len(self.anns_trail) - 1 - frame_i)
                if self._image_meta is not None:
                    frame_anns = [ann.inverse_transform(self._image_meta) for ann in frame_anns]
                self.annotation_painter.annotations(ax, frame_anns, alpha=alpha)



import logging
import json
import zipfile

import numpy as np

from .base import Base

try:
    import pycocotools.coco
    from pycocotools.cocoeval import COCOeval
    # monkey patch for Python 3 compat
    pycocotools.coco.unicode = str
except ImportError:
    COCOeval = None

LOG = logging.getLogger(__name__)


if COCOeval is not None:
    # MonkeyPatch for CrowdPose (or any dataset where the ground truth does not
    # include 'area'):
    # The evaluate() function will call _prepare().
    # However, after _prepare(), we need to add an 'area' to all ground
    # truth instances if not already present based on bbox.
    COCOeval._original_prepare = COCOeval._prepare  # pylint: disable=protected-access

    def new_prepare(instance):
        instance._original_prepare()  # pylint: disable=protected-access
        for gts in instance._gts.values():  # pylint: disable=protected-access
            for gt in gts:
                if 'area' not in gt:
                    gt['area'] = gt['bbox'][2] * gt['bbox'][3]

    COCOeval._prepare = new_prepare  # pylint: disable=protected-access


class Coco(Base):
    text_labels_keypoints = ['AP', 'AP0.5', 'AP0.75', 'APM', 'APL',
                             'AR', 'AR0.5', 'AR0.75', 'ARM', 'ARL']
    text_labels_bbox = ['AP', 'AP0.5', 'AP0.75', 'APS', 'APM', 'APL',
                        'ART1', 'ART10', 'AR', 'ARS', 'ARM', 'ARL']

    def __init__(self, coco, *,
                 max_per_image=20,
                 category_ids=None,
                 iou_type='keypoints',
                 small_threshold=0.0,
                 keypoint_oks_sigmas=None):
        super().__init__()

        if category_ids is None:
            category_ids = [1]

        self.coco = coco
        self.max_per_image = max_per_image
        self.category_ids = category_ids
        self.iou_type = iou_type
        self.small_threshold = small_threshold
        self.keypoint_oks_sigmas = keypoint_oks_sigmas

        self.predictions = []
        self.image_ids = []
        self.eval = None

        if self.iou_type == 'keypoints':
            self.text_labels = self.text_labels_keypoints
        elif self.iou_type == 'bbox':
            self.text_labels = self.text_labels_bbox
        else:
            LOG.warning('Unknown iou type "%s". Specify text_labels yourself.', self.iou_type)

        LOG.debug('max = %d, category ids = %s, iou_type = %s',
                  self.max_per_image, self.category_ids, self.iou_type)

    def _stats(self, predictions=None, image_ids=None):
        # from pycocotools.cocoeval import COCOeval
        if predictions is None:
            predictions = self.predictions
        if image_ids is None:
            image_ids = self.image_ids

        coco_eval = self.coco.loadRes(predictions)

        self.eval = COCOeval(self.coco, coco_eval, iouType=self.iou_type)
        LOG.info('cat_ids: %s', self.category_ids)
        if self.category_ids:
            self.eval.params.catIds = self.category_ids
        if self.keypoint_oks_sigmas is not None:
            self.eval.params.kpt_oks_sigmas = np.asarray(self.keypoint_oks_sigmas)

        if image_ids is not None:
            print('image ids', image_ids)
            self.eval.params.imgIds = image_ids
        self.eval.evaluate()
        self.eval.accumulate()
        self.eval.summarize()
        return self.eval.stats

    # pylint: disable=unused-argument
    def accumulate(self, predictions, image_meta, *, ground_truth=None):
        image_id = int(image_meta['image_id'])
        self.image_ids.append(image_id)

        if self.small_threshold:
            predictions = [pred for pred in predictions
                           if pred.scale(v_th=0.01) >= self.small_threshold]
        if len(predictions) > self.max_per_image:
            predictions = predictions[:self.max_per_image]

        image_annotations = []
        for pred in predictions:
            pred_data = pred.json_data()
            pred_data['image_id'] = image_id
            pred_data = {
                k: v for k, v in pred_data.items()
                if k in ('category_id', 'score', 'keypoints', 'bbox', 'image_id')
            }
            image_annotations.append(pred_data)

        # force at least one annotation per image (for pycocotools)
        if not image_annotations:
            n_keypoints = (len(self.keypoint_oks_sigmas)
                           if self.keypoint_oks_sigmas is not None else 17)
            image_annotations.append({
                'image_id': image_id,
                'category_id': 1,
                'keypoints': np.zeros((n_keypoints * 3,)).tolist(),
                'bbox': [0, 0, 1, 1],
                'score': 0.001,
            })

        if LOG.getEffectiveLevel() == logging.DEBUG:
            self._stats(image_annotations, [image_id])
            LOG.debug(image_meta)

        self.predictions += image_annotations

    def write_predictions(self, filename, *, additional_data=None):
        predictions = [
            {k: v for k, v in annotation.items()
             if k in ('image_id', 'category_id', 'keypoints', 'score')}
            for annotation in self.predictions
        ]
        with open(filename + '.pred.json', 'w') as f:
            json.dump(predictions, f)
        LOG.info('wrote %s.pred.json', filename)
        with zipfile.ZipFile(filename + '.zip', 'w') as myzip:
            myzip.write(filename + '.pred.json', arcname='predictions.json')
        LOG.info('wrote %s.zip', filename)

        if additional_data:
            with open(filename + '.pred_meta.json', 'w') as f:
                json.dump(additional_data, f)
            LOG.info('wrote %s.pred_meta.json', filename)

    def stats(self):
        data = {
            'stats': self._stats().tolist(),
            'text_labels': self.text_labels,
        }

        return data



class SingleImage:
    """Helper class for encoders on datasets with image pairs."""

    def __init__(self, wrapped):
        self.wrapped = wrapped

    def __call__(self, images, anns, metas):
        return self.wrapped(images[0], anns[0], metas[0])

    def __repr__(self):
        return __class__.__module__ + '.' + __class__.__name__ + '(' + repr(self.wrapped) + ')'

    @property
    def meta(self):
        return self.wrapped.meta



import openpifpaf

from . import apollo_kp


def register():
    openpifpaf.DATAMODULES['apollo'] = apollo_kp.ApolloKp
    openpifpaf.CHECKPOINT_URLS['shufflenetv2k16-apollo-24'] = \
        "http://github.com/DuncanZauss/openpifpaf_assets/releases/" \
        "download/v0.1.0/shufflenetv2k16-201113-135121-apollo.pkl.epoch290"
    openpifpaf.CHECKPOINT_URLS['shufflenetv2k16-apollo-66'] = \
        "http://github.com/DuncanZauss/openpifpaf_assets/releases/" \
        "download/v0.1.0/sk16_apollo_66kp.pkl"
    openpifpaf.CHECKPOINT_URLS['shufflenetv2k30-apollo-66'] = \
        "http://github.com/DuncanZauss/openpifpaf_assets/releases/" \
        "download/v0.1.0/sk30_apollo_66kp.pkl"



"""Benchmark."""

import argparse
from collections import namedtuple
import datetime
import logging
import os

import openpifpaf.benchmark

LOG = logging.getLogger(__name__)


DEFAULT_CHECKPOINTS = [
    'tshufflenetv2k16',
]


class CustomFormatter(argparse.ArgumentDefaultsHelpFormatter,
                      argparse.RawDescriptionHelpFormatter):
    pass


def cli():
    parser = argparse.ArgumentParser(
        prog='python3 -m openpifpaf.benchmark',
        description=__doc__,
        formatter_class=CustomFormatter,
    )
    parser.add_argument('--version', action='version',
                        version='OpenPifPaf {version}'.format(version=openpifpaf.__version__))

    parser.add_argument('--output', default=None,
                        help='output file name')
    parser.add_argument('--checkpoints', default=DEFAULT_CHECKPOINTS, nargs='+',
                        help='checkpoints to evaluate')
    parser.add_argument('--crowdpose', default=False, action='store_true')
    parser.add_argument('--ablation-1', default=False, action='store_true')
    parser.add_argument('--ablation-2', default=False, action='store_true')
    parser.add_argument('--ablation-3', default=False, action='store_true')
    parser.add_argument('--ablation-4', default=False, action='store_true')
    parser.add_argument('--ablation-5', default=False, action='store_true')
    group = parser.add_argument_group('logging')
    group.add_argument('--debug', default=False, action='store_true',
                       help='print debug messages')
    args, eval_args = parser.parse_known_args()

    logging.basicConfig(level=logging.INFO if not args.debug else logging.DEBUG)

    # default eval_args
    if not eval_args:
        eval_args = ['--loader-workers=2']

    # default loader workers
    if not any(l.startswith('--loader-workers') for l in eval_args):
        LOG.info('adding "--loader-workers=2" to the argument list')
        eval_args.append('--loader-workers=2')

    # default dataset
    if not any(l.startswith('--dataset') for l in eval_args):
        if args.crowdpose:
            LOG.info('adding "--dataset=crowdpose" to the argument list')
            eval_args.append('--dataset=crowdpose')
            if not any(l.startswith('--force-complete-pose') for l in eval_args):
                LOG.info('adding "--force-complete-pose" to the argument list')
                eval_args.append('--force-complete-pose')
            if not any(l.startswith('--seed-threshold') for l in eval_args):
                LOG.info('adding "--seed-threshold=0.2" to the argument list')
                eval_args.append('--seed-threshold=0.2')
            if not any(l.startswith('--crowdpose-eval-test') for l in eval_args):
                LOG.info('adding "--crowdpose-eval-test" to the argument list')
                eval_args.append('--crowdpose-eval-test')
            if not any(l.startswith('--decoder') for l in eval_args):
                LOG.info('adding "--decoder=cifcaf:0" to the argument list')
                eval_args.append('--decoder=cifcaf:0')
        else:
            LOG.info('adding "--dataset=posetrack2018" to the argument list')
            eval_args.append('--dataset=posetrack2018')
            if not any(l.startswith('--write-predictions') for l in eval_args):
                LOG.info('adding "--write-predictions" to the argument list')
                eval_args.append('--write-predictions')
            if not any(l.startswith('--decoder') for l in eval_args):
                LOG.info('adding "--decoder=trackingpose:0" to the argument list')
                eval_args.append('--decoder=trackingpose:0')

    # generate a default output filename
    if args.output is None:
        now = datetime.datetime.now().strftime('%y%m%d-%H%M%S')
        args.output = 'outputs/benchmark-{}/'.format(now)
        os.makedirs(args.output)

    return args, eval_args


def main():
    args, eval_args = cli()
    Ablation = namedtuple('Ablation', ['suffix', 'args'])
    ablations = [Ablation('', eval_args)]

    if args.crowdpose:
        assert all('crowdpose' in c for c in args.checkpoints)
        ablations += [
            Ablation('.easy', eval_args + ['--crowdpose-index=easy']),
            Ablation('.medium', eval_args + ['--crowdpose-index=medium']),
            Ablation('.hard', eval_args + ['--crowdpose-index=hard']),
        ]

    if args.ablation_1:
        ablations += [
            Ablation('.greedy', eval_args + ['--greedy']),
            Ablation('.no-reverse', eval_args + ['--no-reverse-match']),
            Ablation('.greedy.no-reverse', eval_args + ['--greedy', '--no-reverse-match']),
            # Ablation('.greedy.dense', eval_args + ['--greedy', '--dense-connections']),
            # Ablation('.dense', eval_args + ['--dense-connections']),
            # Ablation('.dense.hierarchy', eval_args + ['--dense-connections=0.1']),
        ]
    if args.ablation_2:
        ablations += [
            Ablation('.nr.nms', eval_args + ['--ablation-cifseeds-no-rescore',
                                             '--ablation-cifseeds-nms',
                                             '--ablation-caf-no-rescore']),
        ]
    if args.ablation_3:
        eval_args_decabl = [
            arg
            for arg in eval_args
            if not arg.startswith(('--instance-threshold=', '--decoder='))
        ]
        ablations += [
            Ablation('.euclidean', eval_args_decabl + ['--decoder=posesimilarity:0',
                                                       '--posesimilarity-distance=euclidean']),
            Ablation('.oks', eval_args_decabl + ['--decoder=posesimilarity:0',
                                                 '--posesimilarity-distance=oks']),
            Ablation('.oks-inflate2', eval_args_decabl + ['--decoder=posesimilarity:0',
                                                          '--posesimilarity-distance=oks',
                                                          '--posesimilarity-oks-inflate=2.0']),
            Ablation('.oks-inflate10', eval_args_decabl + ['--decoder=posesimilarity:0',
                                                           '--posesimilarity-distance=oks',
                                                           '--posesimilarity-oks-inflate=10.0']),
        ]
    if args.ablation_4:
        ablations += [
            Ablation('.w513', eval_args + ['--posetrack-eval-long-edge=513']),
            Ablation('.w641', eval_args + ['--posetrack-eval-long-edge=641']),
            Ablation('.w1201', eval_args + ['--posetrack-eval-long-edge=1201']),
        ]
    if args.ablation_5:
        ablations += [
            Ablation('.recovery', eval_args + ['--trackingpose-track-recovery']),
        ]

    configs = [
        openpifpaf.benchmark.Config(checkpoint, ablation.suffix, ablation.args)
        for checkpoint in args.checkpoints
        for ablation in ablations
    ]
    openpifpaf.benchmark.Benchmark(
        configs,
        args.output,
        reference_config=configs[0] if len(args.checkpoints) == 1 and not args.crowdpose else None,
        stat_scale=100.0 if args.crowdpose else 1.0,
    ).run()


if __name__ == '__main__':
    main()



import cProfile
import io
import logging
import pstats

import torch

LOG = logging.getLogger(__name__)


class Profiler:
    def __init__(self, function_to_profile, *, profile=None, out_name=None):
        if profile is None:
            profile = cProfile.Profile()

        self.function_to_profile = function_to_profile
        self.profile = profile
        self.out_name = out_name

    def __call__(self, *args, **kwargs):
        self.profile.enable()

        result = self.function_to_profile(*args, **kwargs)

        self.profile.disable()
        iostream = io.StringIO()
        ps = pstats.Stats(self.profile, stream=iostream)
        ps = ps.sort_stats('tottime')
        ps.print_stats()
        if self.out_name:
            LOG.info('writing profile file %s', self.out_name)
            ps.dump_stats(self.out_name)
        print(iostream.getvalue())

        return result


class TorchProfiler:
    trace_counter = 0
    table_sort_by = 'self_cpu_time_total'

    def __init__(self, function_to_profile, *, out_name=None):
        if not out_name:
            out_name = 'torchprofiler_chrome_trace.json'

        self.function_to_profile = function_to_profile
        self.out_name = out_name

    def __call__(self, *args, **kwargs):
        with torch.profiler.profile(with_stack=True) as prof:
            result = self.function_to_profile(*args, **kwargs)
        print(prof.key_averages().table(sort_by=self.table_sort_by, row_limit=-1))

        self.__class__.trace_counter += 1
        tracefilename = '{}.{}.json'.format(
            self.out_name.replace('.json', '').replace('.prof', ''),
            self.trace_counter,
        )
        LOG.info('writing trace file %s', tracefilename)
        prof.export_chrome_trace(tracefilename)

        return result



"""
Copyright (c) by Duo Li licensed under the MIT license. [see docs/LICENSE.EFFNETV2 for details]

Creates a EfficientNetV2 Model as defined in:
Mingxing Tan, Quoc V. Le. (2021).
EfficientNetV2: Smaller Models and Faster Training
arXiv preprint arXiv:2104.00298.
"""

import math
from torch import nn


def _make_divisible(v, divisor, min_value=None):
    """
    This function is taken from the original tf repo.
    It ensures that all layers have a channel number that is divisible by 8
    It can be seen here:
    https://github.com/tensorflow/models/blob/master/research/slim/nets/mobilenet/mobilenet.py
    :param v:
    :param divisor:
    :param min_value:
    :return:
    """
    if min_value is None:
        min_value = divisor
    new_v = max(min_value, int(v + divisor / 2) // divisor * divisor)
    # Make sure that round down does not go down by more than 10%.
    if new_v < 0.9 * v:
        new_v += divisor
    return new_v


# SiLU (Swish) activation function
SiLU = nn.SiLU


class SELayer(nn.Module):
    def __init__(self, inp, oup, reduction=4):
        super().__init__()
        self.avg_pool = nn.AdaptiveAvgPool2d(1)
        self.fully_c = nn.Sequential(
            nn.Linear(oup, _make_divisible(inp // reduction, 8)),
            SiLU(),
            nn.Linear(_make_divisible(inp // reduction, 8), oup),
            nn.Sigmoid()
        )

    def forward(self, x):
        b, c, _, _ = x.size()
        y = self.avg_pool(x).view(b, c)
        y = self.fully_c(y).view(b, c, 1, 1)
        return x * y


def conv_3x3_bn(inp, oup, stride):
    return nn.Sequential(
        nn.Conv2d(inp, oup, 3, stride, 1, bias=False),
        nn.BatchNorm2d(oup),
        SiLU()
    )


def conv_1x1_bn(inp, oup):
    return nn.Sequential(
        nn.Conv2d(inp, oup, 1, 1, 0, bias=False),
        nn.BatchNorm2d(oup),
        SiLU()
    )


class MBConv(nn.Module):

    def __init__(self, inp, oup, stride, expand_ratio, use_se):
        super().__init__()
        assert stride in [-1, 1, 2]

        hidden_dim = round(inp * expand_ratio)
        self.identity = stride == 1 and inp == oup
        if use_se:
            self.conv = nn.Sequential(
                # Paper Figure 2 MBConv
                # pw
                # inp size, outp size, kernel size, stride, padding
                nn.Conv2d(inp, hidden_dim, 1, 1, 0, bias=False),
                nn.BatchNorm2d(hidden_dim),
                SiLU(),
                # dw
                nn.Conv2d(hidden_dim, hidden_dim, 3, stride, 1, groups=hidden_dim, bias=False) if \
                stride != -1 else nn.Conv2d(hidden_dim, hidden_dim, 3, stride=1, padding=2,
                                            groups=hidden_dim, dilation=2, bias=False),

                nn.BatchNorm2d(hidden_dim),
                SiLU(),
                SELayer(inp, hidden_dim),
                # pw-linear
                nn.Conv2d(hidden_dim, oup, 1, 1, 0, bias=False),
                nn.BatchNorm2d(oup),
            )
        else:
            self.conv = nn.Sequential(
                # fused
                nn.Conv2d(inp, hidden_dim, 3, stride, 1, bias=False),
                nn.BatchNorm2d(hidden_dim),
                SiLU(),
                # pw-linear
                nn.Conv2d(hidden_dim, oup, 1, 1, 0, bias=False),
                nn.BatchNorm2d(oup),
            )

    def forward(self, x):
        if self.identity:
            return x + self.conv(x)
        return self.conv(x)


class EffNetV2(nn.Module):
    def __init__(self, cfgs, width_mult=1.):
        super().__init__()
        self.cfgs = cfgs

        # building first layer
        input_channel = _make_divisible(24 * width_mult, 8)
        layers = [conv_3x3_bn(3, input_channel, 2)]
        # building inverted residual blocks
        block = MBConv
        for t, c, n, s, use_se in self.cfgs:
            output_channel = _make_divisible(c * width_mult, 8)
            for i in range(n):
                layers.append(block(input_channel, output_channel, s if i == 0 else 1, t, use_se))
                input_channel = output_channel
        self.features = nn.Sequential(*layers)
        self.output_channel = output_channel
        # building last several layers
        # output_channel = _make_divisible(1792 * width_mult, 8) if width_mult > 1.0 else 1792
        # self.conv = conv_1x1_bn(input_channel, output_channel)
        # self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        # self.classifier = nn.Linear(output_channel, num_classes)

        self._initialize_weights()

    def forward(self, x):
        x = self.features(x)
        # x = self.conv(x)
        # x = self.avgpool(x)
        # x = x.view(x.size(0), -1)
        # x = self.classifier(x)
        return x

    def _initialize_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
                if m.bias is not None:
                    m.bias.data.zero_()
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()
            elif isinstance(m, nn.Linear):
                m.weight.data.normal_(0, 0.001)
                m.bias.data.zero_()



"""An open implementation of PifPaf."""

# pylint: disable=wrong-import-position

from . import _version
__version__ = _version.get_versions()['version']

# register ops first
from . import cpp_extension
cpp_extension.register_ops()

from .annotation import Annotation, AnnotationDet
from .configurable import Configurable
from .predictor import Predictor
from .signal import Signal
from . import datasets
from . import decoder
from . import encoder
from . import headmeta
from . import logger
from . import metric
from . import network
from . import optimize
from . import plugin
from . import visualizer

from .datasets import DATAMODULES
from .decoder import DECODERS
from .network.factory import (
    BASE_FACTORIES,
    BASE_TYPES,
    CHECKPOINT_URLS,
    HEADS,
    PRETRAINED_UNAVAILABLE,
)
from .network.losses.factory import LOSSES, LOSS_COMPONENTS
from .network.model_migration import MODEL_MIGRATION
from .show.annotation_painter import PAINTERS

# load plugins last
plugin.register()



from collections import defaultdict
import copy
import logging
import os

import torch.utils.data
from PIL import Image

import openpifpaf


LOG = logging.getLogger(__name__)
STAT_LOG = logging.getLogger(__name__.replace('openpifpaf.', 'openpifpaf.stats.'))


class CocoDataset(torch.utils.data.Dataset):
    """`MS Coco Detection <http://mscoco.org/dataset/#detections-challenge2016>`_ Dataset.

    Args:
        image_dir (string): Root directory where images are downloaded to.
        ann_file (string): Path to json annotation file.
    """

    def __init__(self, image_dir, ann_file, *,
                 preprocess=None, min_kp_anns=0,
                 category_ids=None,
                 annotation_filter=False):
        super().__init__()

        if category_ids is None:
            category_ids = []

        from pycocotools.coco import COCO  # pylint: disable=import-outside-toplevel
        self.image_dir = image_dir
        self.coco = COCO(ann_file)

        self.category_ids = category_ids

        self.ids = self.coco.getImgIds(catIds=self.category_ids)
        if annotation_filter:
            self.filter_for_annotations(min_kp_anns=min_kp_anns)
        elif min_kp_anns:
            raise Exception('only set min_kp_anns with annotation_filter')
        LOG.info('Images: %d', len(self.ids))

        self.preprocess = preprocess or openpifpaf.transforms.EVAL_TRANSFORM

    def filter_for_annotations(self, *, min_kp_anns=0):
        LOG.info('filter for annotations (min kp=%d) ...', min_kp_anns)

        def filter_image(image_id):
            ann_ids = self.coco.getAnnIds(imgIds=image_id, catIds=self.category_ids)
            anns = self.coco.loadAnns(ann_ids)
            anns = [ann for ann in anns if not ann.get('iscrowd')]
            if not anns:
                return False
            kp_anns = [ann for ann in anns
                       if 'keypoints' in ann and any(v > 0.0 for v in ann['keypoints'][2::3])]
            return len(kp_anns) >= min_kp_anns

        self.ids = [image_id for image_id in self.ids if filter_image(image_id)]
        LOG.info('... done.')

    def class_aware_sample_weights(self, max_multiple=10.0):
        """Class aware sampling.

        To be used with PyTorch's WeightedRandomSampler.

        Reference: Solution for Large-Scale Hierarchical Object Detection
        Datasets with Incomplete Annotation and Data Imbalance
        Yuan Gao, Xingyuan Bu, Yang Hu, Hui Shen, Ti Bai, Xubin Li and Shilei Wen
        """
        ann_ids = self.coco.getAnnIds(imgIds=self.ids, catIds=self.category_ids)
        anns = self.coco.loadAnns(ann_ids)

        category_image_counts = defaultdict(int)
        image_categories = defaultdict(set)
        for ann in anns:
            if ann['iscrowd']:
                continue
            image = ann['image_id']
            category = ann['category_id']
            if category in image_categories[image]:
                continue
            image_categories[image].add(category)
            category_image_counts[category] += 1

        weights = [
            sum(
                1.0 / category_image_counts[category_id]
                for category_id in image_categories[image_id]
            )
            for image_id in self.ids
        ]
        min_w = min(weights)
        LOG.debug('Class Aware Sampling: minW = %f, maxW = %f', min_w, max(weights))
        max_w = min_w * max_multiple
        weights = [min(w, max_w) for w in weights]
        LOG.debug('Class Aware Sampling: minW = %f, maxW = %f', min_w, max(weights))

        return weights

    def __getitem__(self, index):
        image_id = self.ids[index]
        ann_ids = self.coco.getAnnIds(imgIds=image_id, catIds=self.category_ids)
        anns = self.coco.loadAnns(ann_ids)
        anns = copy.deepcopy(anns)

        image_info = self.coco.loadImgs(image_id)[0]
        LOG.debug(image_info)
        local_file_path = os.path.join(self.image_dir, image_info['file_name'])
        with open(local_file_path, 'rb') as f:
            image = Image.open(f).convert('RGB')

        meta = {
            'dataset_index': index,
            'image_id': image_id,
            'file_name': image_info['file_name'],
            'local_file_path': local_file_path,
        }

        if 'flickr_url' in image_info:
            _, flickr_file_name = image_info['flickr_url'].rsplit('/', maxsplit=1)
            flickr_id, _ = flickr_file_name.split('_', maxsplit=1)
            meta['flickr_full_page'] = 'http://flickr.com/photo.gne?id={}'.format(flickr_id)

        # preprocess image and annotations
        image, anns, meta = self.preprocess(image, anns, meta)

        LOG.debug(meta)

        # TODO: convert into transform
        # # log stats
        # for ann in anns:
        #     if getattr(ann, 'iscrowd', False):
        #         continue
        #     if not np.any(ann['keypoints'][:, 2] > 0.0):
        #         continue
        #     STAT_LOG.debug({'bbox': [int(v) for v in ann['bbox']]})

        return image, anns, meta

    def __len__(self):
        return len(self.ids)



import numpy as np

import openpifpaf

from .posetrack2018 import KEYPOINTS, SKELETON, SIGMAS, UPRIGHT_POSE


def main():
    openpifpaf.show.KeypointPainter.show_joint_scales = True
    openpifpaf.show.KeypointPainter.line_width = 6
    openpifpaf.show.KeypointPainter.monocolor_connections = False
    keypoint_painter = openpifpaf.show.KeypointPainter()

    scale = np.sqrt(
        (np.max(UPRIGHT_POSE[:, 0]) - np.min(UPRIGHT_POSE[:, 0]))
        * (np.max(UPRIGHT_POSE[:, 1]) - np.min(UPRIGHT_POSE[:, 1]))
    )

    ann_posetrack = openpifpaf.Annotation(KEYPOINTS, SKELETON)
    ann_posetrack.set(UPRIGHT_POSE, np.array(SIGMAS) * scale, fixed_score='')
    with openpifpaf.show.Canvas.annotation(
            ann_posetrack, filename='docs/skeleton_posetrack.png') as ax:
        keypoint_painter.annotation(ax, ann_posetrack)

    upright_pose_2tracking = np.concatenate([
        UPRIGHT_POSE,
        0.9 * UPRIGHT_POSE + np.array([-1.5, 1.5, 0.0]),
    ])
    sigmas_2tracking = np.concatenate([np.array(SIGMAS) * scale, 0.8 * np.array(SIGMAS) * scale])
    tracking2_skeleton = np.concatenate([
        np.array(SKELETON) + 17,
        np.array([(j, j + 17) for j in range(1, 18)]),
        np.array(SKELETON),
    ])
    ann_tracking = openpifpaf.Annotation(KEYPOINTS + KEYPOINTS, tracking2_skeleton)
    ann_tracking.set(
        upright_pose_2tracking,
        sigmas_2tracking,
        fixed_score='',
    )
    with openpifpaf.show.Canvas.annotation(
            ann_tracking, filename='docs/skeleton_tracking.png') as ax:
        keypoint_painter.annotation(ax, ann_tracking)

    tracking_skeleton_forward = np.concatenate([
        np.array([(j, j + 17) for j in range(1, 18)]),
        np.array(SKELETON),
    ])
    ann_tracking_forward = openpifpaf.Annotation(KEYPOINTS + KEYPOINTS, tracking_skeleton_forward)
    ann_tracking_forward.set(
        upright_pose_2tracking,
        sigmas_2tracking,
        fixed_score='',
    )
    with openpifpaf.show.Canvas.annotation(
            ann_tracking_forward, filename='docs/skeleton_tracking_forward.png') as ax:
        keypoint_painter.annotation(ax, ann_tracking_forward)

    # COCO
    coco_keypoints = openpifpaf.plugins.coco.constants.COCO_KEYPOINTS
    coco_skeleton = openpifpaf.plugins.coco.constants.COCO_PERSON_SKELETON
    coco_skeleton_forward = np.concatenate([
        np.array([(j, j + 17) for j in range(1, 18)]),
        np.array(coco_skeleton),
    ])
    coco_upright_pose_2tracking = np.concatenate([
        openpifpaf.plugins.coco.constants.COCO_UPRIGHT_POSE,
        0.9 * openpifpaf.plugins.coco.constants.COCO_UPRIGHT_POSE + np.array([-1.5, 1.5, 0.0]),
    ])
    coco_sigmas_2tracking = np.concatenate([
        np.array(openpifpaf.plugins.coco.constants.COCO_PERSON_SIGMAS) * scale,
        0.8 * np.array(openpifpaf.plugins.coco.constants.COCO_PERSON_SIGMAS) * scale,
    ])
    ann_coco_tracking_forward = openpifpaf.Annotation(
        coco_keypoints + coco_keypoints, coco_skeleton_forward)
    ann_coco_tracking_forward.set(
        coco_upright_pose_2tracking,
        coco_sigmas_2tracking,
        fixed_score='',
    )
    with openpifpaf.show.Canvas.annotation(
            ann_coco_tracking_forward, filename='docs/coco_skeleton_forward.png') as ax:
        keypoint_painter.annotation(ax, ann_coco_tracking_forward)

    # overview
    with openpifpaf.show.Canvas.blank(
            'docs/skeleton_overview.png', figsize=(12, 6), ncols=4) as (ax1, ax2, ax3, ax4):
        ax1.set_axis_off()
        ax1.set_aspect('equal')
        keypoint_painter.annotation(ax1, ann_posetrack)

        ax2.set_axis_off()
        ax2.set_aspect('equal')
        keypoint_painter.annotation(ax2, ann_tracking)

        ax3.set_axis_off()
        ax3.set_aspect('equal')
        keypoint_painter.annotation(ax3, ann_tracking_forward)

        ax4.set_axis_off()
        ax4.set_aspect('equal')
        keypoint_painter.annotation(ax4, ann_coco_tracking_forward)


if __name__ == '__main__':
    main()



import logging

import numpy as np

LOG = logging.getLogger(__name__)


class AnnRescaler():
    suppress_selfhidden = True
    suppress_invisible = False
    suppress_collision = False

    def __init__(self, stride, pose=None):
        self.stride = stride
        self.pose = pose

        self.pose_total_area = None
        self.pose_45 = None
        self.pose_45_total_area = None
        if pose is not None:
            self.pose_total_area = (
                (np.max(self.pose[:, 0]) - np.min(self.pose[:, 0]))
                * (np.max(self.pose[:, 1]) - np.min(self.pose[:, 1]))
            )

            # rotate the davinci pose by 45 degrees
            c, s = np.cos(np.deg2rad(45)), np.sin(np.deg2rad(45))
            rotate = np.array(((c, -s), (s, c)))
            self.pose_45 = np.copy(self.pose)
            self.pose_45[:, :2] = np.einsum('ij,kj->ki', rotate, self.pose_45[:, :2])
            self.pose_45_total_area = (
                (np.max(self.pose_45[:, 0]) - np.min(self.pose_45[:, 0]))
                * (np.max(self.pose_45[:, 1]) - np.min(self.pose_45[:, 1]))
            )

    def valid_area(self, meta):
        if 'valid_area' not in meta:
            return None

        return (
            meta['valid_area'][0] / self.stride,
            meta['valid_area'][1] / self.stride,
            meta['valid_area'][2] / self.stride,
            meta['valid_area'][3] / self.stride,
        )

    @staticmethod
    def suppress_collision_(keypoint_sets_bbox):
        for p_i, (kps_p, bbox_p) in enumerate(keypoint_sets_bbox[:-1]):
            for kps_s, bbox_s in keypoint_sets_bbox[p_i + 1:]:
                d_th = 0.2 * max(bbox_p[2], bbox_p[3], bbox_s[2], bbox_s[3])
                d_th = max(16.0, d_th)
                diff = np.abs(kps_p[:, :2] - kps_s[:, :2])
                collision = (
                    (kps_p[:, 2] > 0.0)
                    & (kps_s[:, 2] > 0.0)
                    & (diff[:, 0] < d_th)
                    & (diff[:, 1] < d_th)
                )
                if np.any(collision):
                    kps_p[collision, 2] = 0.0
                    kps_s[collision, 2] = 0.0

    @staticmethod
    def suppress_selfhidden_(keypoint_sets):
        for kpi in range(len(keypoint_sets[0])):
            all_xyv = sorted([keypoints[kpi] for keypoints in keypoint_sets],
                             key=lambda xyv: xyv[2], reverse=True)
            for i, xyv in enumerate(all_xyv[1:], start=1):
                if xyv[2] > 1.0:  # is visible
                    continue
                if xyv[2] < 1.0:  # does not exist
                    break
                for prev_xyv in all_xyv[:i]:
                    if prev_xyv[2] <= 1.0:  # do not suppress if both hidden
                        break
                    if np.abs(prev_xyv[0] - xyv[0]) > 32.0 \
                       or np.abs(prev_xyv[1] - xyv[1]) > 32.0:
                        continue
                    LOG.debug('suppressing %s for %s (kp %d)', xyv, prev_xyv, i)
                    xyv[2] = 0.0
                    break  # only need to suppress a keypoint once

    def keypoint_sets(self, anns):
        """Ignore annotations of crowds."""
        keypoint_sets_bbox = [(np.copy(ann['keypoints']), ann['bbox'])
                              for ann in anns if not ann['iscrowd']]
        if not keypoint_sets_bbox:
            return []

        if self.suppress_collision:
            self.suppress_collision_(keypoint_sets_bbox)
        keypoint_sets = [kps for kps, _ in keypoint_sets_bbox]

        if self.suppress_invisible:
            for kps in keypoint_sets:
                kps[kps[:, 2] < 2.0, 2] = 0.0
        elif self.suppress_selfhidden:
            self.suppress_selfhidden_(keypoint_sets)

        for keypoints in keypoint_sets:
            keypoints[:, :2] /= self.stride
        return keypoint_sets

    def bg_mask(self, anns, width_height, *, crowd_margin):
        """Create background mask taking crowd annotations into account."""
        mask = np.ones((
            (width_height[1] - 1) // self.stride + 1,
            (width_height[0] - 1) // self.stride + 1,
        ), dtype=np.bool)
        for ann in anns:
            if not ann['iscrowd']:
                valid_keypoints = 'keypoints' in ann and np.any(ann['keypoints'][:, 2] > 0)
                if valid_keypoints:
                    continue

            if 'mask' not in ann:
                bb = ann['bbox'].copy()
                bb /= self.stride
                bb[2:] += bb[:2]  # convert width and height to x2 and y2

                # left top
                left = np.clip(int(bb[0] - crowd_margin), 0, mask.shape[1] - 1)
                top = np.clip(int(bb[1] - crowd_margin), 0, mask.shape[0] - 1)

                # right bottom
                # ceil: to round up
                # +1: because mask upper limit is exclusive
                right = np.clip(int(np.ceil(bb[2] + crowd_margin)) + 1,
                                left + 1, mask.shape[1])
                bottom = np.clip(int(np.ceil(bb[3] + crowd_margin)) + 1,
                                 top + 1, mask.shape[0])

                mask[top:bottom, left:right] = 0
                continue

            assert False  # because code below is not tested
            mask[ann['mask'][::self.stride, ::self.stride]] = 0

        return mask

    def scale(self, keypoints):
        visible = keypoints[:, 2] > 0
        if np.sum(visible) < 3:
            return np.nan

        area = (
            (np.max(keypoints[visible, 0]) - np.min(keypoints[visible, 0]))
            * (np.max(keypoints[visible, 1]) - np.min(keypoints[visible, 1]))
        )
        factor = 1.0

        if self.pose is not None:
            area_ref = (
                (np.max(self.pose[visible, 0]) - np.min(self.pose[visible, 0]))
                * (np.max(self.pose[visible, 1]) - np.min(self.pose[visible, 1]))
            )
            area_ref_45 = (
                (np.max(self.pose_45[visible, 0]) - np.min(self.pose_45[visible, 0]))
                * (np.max(self.pose_45[visible, 1]) - np.min(self.pose_45[visible, 1]))
            )

            factor = np.sqrt(min(
                self.pose_total_area / area_ref if area_ref > 0.1 else np.inf,
                self.pose_45_total_area / area_ref_45 if area_ref_45 > 0.1 else np.inf,
            ))
            if np.isinf(factor):
                return np.nan

        factor_clipped = min(5.0, factor)
        scale = np.sqrt(area) * factor_clipped
        if scale < 0.1:
            scale = np.nan

        LOG.debug('instance scale = %.3f (factor = %.2f, clipped factor = %.2f)',
                  scale, factor, factor_clipped)
        return scale


class AnnRescalerDet():
    def __init__(self, stride, n_categories):
        self.stride = stride
        self.n_categories = n_categories

    def valid_area(self, meta):
        if 'valid_area' not in meta:
            return None

        return (
            meta['valid_area'][0] / self.stride,
            meta['valid_area'][1] / self.stride,
            meta['valid_area'][2] / self.stride,
            meta['valid_area'][3] / self.stride,
        )

    def detections(self, anns):
        category_bboxes = [(ann['category_id'], ann['bbox'] / self.stride)
                           for ann in anns if not ann['iscrowd']]
        return category_bboxes

    def bg_mask(self, anns, width_height, *, crowd_margin):
        """Create background mask taking crowd annotations into account."""
        mask = np.ones((
            self.n_categories,
            (width_height[1] - 1) // self.stride + 1,
            (width_height[0] - 1) // self.stride + 1,
        ), dtype=np.bool)
        for ann in anns:
            if not ann['iscrowd']:
                continue

            if 'mask' not in ann:
                field_i = ann['category_id'] - 1
                bb = ann['bbox'].copy()
                bb /= self.stride
                bb[2:] += bb[:2]  # convert width and height to x2 and y2
                left = np.clip(int(bb[0] - crowd_margin), 0, mask.shape[1] - 1)
                top = np.clip(int(bb[1] - crowd_margin), 0, mask.shape[0] - 1)
                right = np.clip(int(np.ceil(bb[2] + crowd_margin)) + 1,
                                left + 1, mask.shape[1])
                bottom = np.clip(int(np.ceil(bb[3] + crowd_margin)) + 1,
                                 top + 1, mask.shape[0])
                mask[field_i, top:bottom, left:right] = 0
                continue

            assert False  # because code below is not tested
            mask[ann['mask'][::self.stride, ::self.stride]] = 0

        return mask


class TrackingAnnRescaler(AnnRescaler):
    def bg_mask(self, anns, width_height, *, crowd_margin):
        """Create background mask taking crowd annotations into account."""
        anns1, anns2 = anns

        mask = np.ones((
            (width_height[1] - 1) // self.stride + 1,
            (width_height[0] - 1) // self.stride + 1,
        ), dtype=np.bool)
        crowd_bbox = [np.inf, np.inf, 0, 0]
        for ann in anns1 + anns2:
            if not ann['iscrowd']:
                valid_keypoints = 'keypoints' in ann and np.any(ann['keypoints'][:, 2] > 0)
                if valid_keypoints:
                    continue

            if 'mask' not in ann:
                bb = ann['bbox'].copy()
                bb /= self.stride
                bb[2:] += bb[:2]  # convert width and height to x2 and y2

                # left top
                left = np.clip(int(bb[0] - crowd_margin), 0, mask.shape[1] - 1)
                top = np.clip(int(bb[1] - crowd_margin), 0, mask.shape[0] - 1)

                # right bottom
                # ceil: to round up
                # +1: because mask upper limit is exclusive
                right = np.clip(int(np.ceil(bb[2] + crowd_margin)) + 1,
                                left + 1, mask.shape[1])
                bottom = np.clip(int(np.ceil(bb[3] + crowd_margin)) + 1,
                                 top + 1, mask.shape[0])

                crowd_bbox[0] = min(crowd_bbox[0], left)
                crowd_bbox[1] = min(crowd_bbox[1], top)
                crowd_bbox[2] = max(crowd_bbox[2], right)
                crowd_bbox[3] = max(crowd_bbox[3], bottom)
                continue

            assert False  # because code below is not tested
            mask[ann['mask'][::self.stride, ::self.stride]] = 0

        if crowd_bbox[1] < crowd_bbox[3] and crowd_bbox[0] < crowd_bbox[2]:
            LOG.debug('crowd_bbox: %s', crowd_bbox)
            mask[crowd_bbox[1]:crowd_bbox[3], crowd_bbox[0]:crowd_bbox[2]] = 0

        return mask

    def keypoint_sets(self, anns):
        """Ignore annotations of crowds."""
        anns1, anns2 = anns

        anns1_by_trackid = {ann['track_id']: ann for ann in anns1}
        keypoint_sets_bbox = [
            (
                np.concatenate((
                    anns1_by_trackid[ann2['track_id']]['keypoints'],
                    ann2['keypoints'],
                ), axis=0),
                ann2['bbox'],
            )
            for ann2 in anns2
            if (not ann2['iscrowd']
                and ann2['track_id'] in anns1_by_trackid)
        ]
        if not keypoint_sets_bbox:
            return []

        if self.suppress_collision:
            self.suppress_collision_(keypoint_sets_bbox)
        keypoint_sets = [kps for kps, _ in keypoint_sets_bbox]

        if self.suppress_invisible:
            for kps in keypoint_sets:
                kps[kps[:, 2] < 2.0, 2] = 0.0

        for keypoints in keypoint_sets:
            keypoints[:, :2] /= self.stride
        return keypoint_sets



import torch

from . import nms

# pylint: disable=invalid-name
CafScored = torch.classes.openpifpaf_decoder_utils.CafScored
CifHr = torch.classes.openpifpaf_decoder_utils.CifHr
CifSeeds = torch.classes.openpifpaf_decoder_utils.CifSeeds
CifDetSeeds = torch.classes.openpifpaf_decoder_utils.CifDetSeeds
Occupancy = torch.classes.openpifpaf_decoder_utils.Occupancy



"""
Interface for custom data.

This module handles datasets and is the class that you need to inherit from for your custom dataset.
This class gives you all the handles so that you can train with a new –dataset=mydataset.
The particular configuration of keypoints and skeleton is specified in the headmeta instances
"""


import argparse
import torch

try:
    from pycocotools.coco import COCO
except ImportError:
    COCO = None

from openpifpaf.datasets import DataModule
from openpifpaf import encoder, headmeta, metric, transforms
from openpifpaf.datasets import collate_images_anns_meta, collate_images_targets_meta
from openpifpaf.plugins.coco import CocoDataset as CocoLoader

from .constants import ANIMAL_KEYPOINTS, ANIMAL_SKELETON, HFLIP, \
    ANIMAL_SIGMAS, ANIMAL_POSE, ANIMAL_CATEGORIES, ANIMAL_SCORE_WEIGHTS


class AnimalKp(DataModule):
    """
    Adapted from the standard CocoKp class to work as external plugin
    """

    train_annotations = 'data-animalpose/annotations/animal_keypoints_20_train.json'
    val_annotations = 'data-animalpose/annotations/animal_keypoints_20_val.json'
    eval_annotations = val_annotations
    train_image_dir = 'data-animalpose/images/train/'
    val_image_dir = 'data-animalpose/images/val/'
    eval_image_dir = val_image_dir

    n_images = None
    square_edge = 513
    extended_scale = False
    orientation_invariant = 0.0
    blur = 0.0
    augmentation = True
    rescale_images = 1.0
    upsample_stride = 1
    min_kp_anns = 1
    b_min = 1  # 1 pixel

    eval_annotation_filter = True
    eval_long_edge = 0  # set to zero to deactivate rescaling
    eval_orientation_invariant = 0.0
    eval_extended_scale = False

    def __init__(self):
        super().__init__()

        cif = headmeta.Cif('cif', 'animal',
                           keypoints=ANIMAL_KEYPOINTS,
                           sigmas=ANIMAL_SIGMAS,
                           pose=ANIMAL_POSE,
                           draw_skeleton=ANIMAL_SKELETON,
                           score_weights=ANIMAL_SCORE_WEIGHTS)
        caf = headmeta.Caf('caf', 'animal',
                           keypoints=ANIMAL_KEYPOINTS,
                           sigmas=ANIMAL_SIGMAS,
                           pose=ANIMAL_POSE,
                           skeleton=ANIMAL_SKELETON)

        cif.upsample_stride = self.upsample_stride
        caf.upsample_stride = self.upsample_stride
        self.head_metas = [cif, caf]

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        group = parser.add_argument_group('data module Animal')

        group.add_argument('--animal-train-annotations',
                           default=cls.train_annotations)
        group.add_argument('--animal-val-annotations',
                           default=cls.val_annotations)
        group.add_argument('--animal-train-image-dir',
                           default=cls.train_image_dir)
        group.add_argument('--animal-val-image-dir',
                           default=cls.val_image_dir)

        group.add_argument('--animal-square-edge',
                           default=cls.square_edge, type=int,
                           help='square edge of input images')
        assert not cls.extended_scale
        group.add_argument('--animal-extended-scale',
                           default=False, action='store_true',
                           help='augment with an extended scale range')
        group.add_argument('--animal-orientation-invariant',
                           default=cls.orientation_invariant, type=float,
                           help='augment with random orientations')
        group.add_argument('--animal-blur',
                           default=cls.blur, type=float,
                           help='augment with blur')
        assert cls.augmentation
        group.add_argument('--animal-no-augmentation',
                           dest='animal_augmentation',
                           default=True, action='store_false',
                           help='do not apply data augmentation')
        group.add_argument('--animal-rescale-images',
                           default=cls.rescale_images, type=float,
                           help='overall rescale factor for images')
        group.add_argument('--animal-upsample',
                           default=cls.upsample_stride, type=int,
                           help='head upsample stride')
        group.add_argument('--animal-min-kp-anns',
                           default=cls.min_kp_anns, type=int,
                           help='filter images with fewer keypoint annotations')
        group.add_argument('--animal-bmin',
                           default=cls.b_min, type=int,
                           help='b minimum in pixels')

        # evaluation  (TO setup directly)
        eval_set_group = group.add_mutually_exclusive_group()
        eval_set_group.add_argument('--animal-eval-test2017', default=False, action='store_true')
        eval_set_group.add_argument('--animal-eval-testdev2017', default=False, action='store_true')

        group.add_argument('--animal-no-eval-annotation-filter',
                           dest='animal_eval_annotation_filter',
                           default=True, action='store_false')
        group.add_argument('--animal-eval-long-edge', default=cls.eval_long_edge, type=int,
                           dest='animal_eval_long_edge', help='set to zero to deactivate rescaling')
        assert not cls.eval_extended_scale
        group.add_argument('--animal-eval-extended-scale', default=False, action='store_true',
                           dest='animal_eval_extended_scale',)
        group.add_argument('--animal-eval-orientation-invariant',
                           default=cls.eval_orientation_invariant, type=float,
                           dest='animal_eval_orientation_invariant')

    @classmethod
    def configure(cls, args: argparse.Namespace):
        # extract global information
        cls.debug = args.debug
        cls.pin_memory = args.pin_memory

        # Animal specific
        cls.train_annotations = args.animal_train_annotations
        cls.val_annotations = args.animal_val_annotations
        cls.train_image_dir = args.animal_train_image_dir
        cls.val_image_dir = args.animal_val_image_dir

        cls.square_edge = args.animal_square_edge
        cls.extended_scale = args.animal_extended_scale
        cls.orientation_invariant = args.animal_orientation_invariant
        cls.blur = args.animal_blur
        cls.augmentation = args.animal_augmentation  # loaded by the dest name
        cls.rescale_images = args.animal_rescale_images
        cls.upsample_stride = args.animal_upsample
        cls.min_kp_anns = args.animal_min_kp_anns
        cls.b_min = args.animal_bmin

        # evaluation
        cls.eval_annotation_filter = args.animal_eval_annotation_filter
        cls.eval_long_edge = args.animal_eval_long_edge
        cls.eval_orientation_invariant = args.animal_eval_orientation_invariant
        cls.eval_extended_scale = args.animal_eval_extended_scale

        if (args.animal_eval_test2017 or args.animal_eval_testdev2017) \
                and not args.write_predictions \
                and not args.debug:
            raise Exception('have to use --write-predictions for this dataset')

    def _preprocess(self):
        encoders = (encoder.Cif(self.head_metas[0], bmin=self.b_min),
                    encoder.Caf(self.head_metas[1], bmin=self.b_min))

        if not self.augmentation:
            return transforms.Compose([
                transforms.NormalizeAnnotations(),
                transforms.RescaleAbsolute(self.square_edge),
                transforms.CenterPad(self.square_edge),
                transforms.EVAL_TRANSFORM,
                transforms.Encoders(encoders),
            ])

        if self.extended_scale:
            rescale_t = transforms.RescaleRelative(
                scale_range=(0.2 * self.rescale_images,
                             2.5 * self.rescale_images),
                power_law=True, stretch_range=(0.75, 1.33))
        else:
            rescale_t = transforms.RescaleRelative(
                scale_range=(0.3 * self.rescale_images,
                             2.0 * self.rescale_images),
                power_law=True, stretch_range=(0.75, 1.33))

        return transforms.Compose([
            transforms.NormalizeAnnotations(),
            transforms.RandomApply(
                transforms.HFlip(ANIMAL_KEYPOINTS, HFLIP), 0.5),
            rescale_t,
            transforms.RandomApply(
                transforms.Blur(), self.blur),
            transforms.RandomChoice(
                [transforms.RotateBy90(),
                 transforms.RotateUniform(30.0)],
                [self.orientation_invariant, 0.4],
            ),
            transforms.Crop(self.square_edge, use_area_of_interest=True),
            transforms.CenterPad(self.square_edge),
            transforms.TRAIN_TRANSFORM,
            transforms.Encoders(encoders),
        ])

    def train_loader(self):
        train_data = CocoLoader(
            image_dir=self.train_image_dir,
            ann_file=self.train_annotations,
            preprocess=self._preprocess(),
            annotation_filter=True,
            min_kp_anns=self.min_kp_anns,
            category_ids=[1],
        )
        return torch.utils.data.DataLoader(
            train_data, batch_size=self.batch_size, shuffle=not self.debug,
            pin_memory=self.pin_memory, num_workers=self.loader_workers, drop_last=True,
            collate_fn=collate_images_targets_meta)

    def val_loader(self):
        val_data = CocoLoader(
            image_dir=self.val_image_dir,
            ann_file=self.val_annotations,
            preprocess=self._preprocess(),
            annotation_filter=True,
            min_kp_anns=self.min_kp_anns,
            category_ids=[1],
        )
        return torch.utils.data.DataLoader(
            val_data, batch_size=self.batch_size, shuffle=False,
            pin_memory=self.pin_memory, num_workers=self.loader_workers, drop_last=True,
            collate_fn=collate_images_targets_meta)

    @classmethod
    def common_eval_preprocess(cls):
        rescale_t = None
        if cls.eval_extended_scale:
            assert cls.eval_long_edge
            rescale_t = [
                transforms.DeterministicEqualChoice([
                    transforms.RescaleAbsolute(cls.eval_long_edge),
                    transforms.RescaleAbsolute((cls.eval_long_edge - 1) // 2 + 1),
                ], salt=1)
            ]
        elif cls.eval_long_edge:
            rescale_t = transforms.RescaleAbsolute(cls.eval_long_edge)

        if cls.batch_size == 1:
            padding_t = transforms.CenterPadTight(16)
        else:
            assert cls.eval_long_edge
            padding_t = transforms.CenterPad(cls.eval_long_edge)

        orientation_t = None
        if cls.eval_orientation_invariant:
            orientation_t = transforms.DeterministicEqualChoice([
                None,
                transforms.RotateBy90(fixed_angle=90),
                transforms.RotateBy90(fixed_angle=180),
                transforms.RotateBy90(fixed_angle=270),
            ], salt=3)

        return [
            transforms.NormalizeAnnotations(),
            rescale_t,
            padding_t,
            orientation_t,
        ]

    def _eval_preprocess(self):
        return transforms.Compose([
            *self.common_eval_preprocess(),
            transforms.ToAnnotations([
                transforms.ToKpAnnotations(
                    ANIMAL_CATEGORIES,
                    keypoints_by_category={1: self.head_metas[0].keypoints},
                    skeleton_by_category={1: self.head_metas[1].skeleton},
                ),
                transforms.ToCrowdAnnotations(ANIMAL_CATEGORIES),
            ]),
            transforms.EVAL_TRANSFORM,
        ])

    def eval_loader(self):
        eval_data = CocoLoader(
            image_dir=self.eval_image_dir,
            ann_file=self.eval_annotations,
            preprocess=self._eval_preprocess(),
            annotation_filter=self.eval_annotation_filter,
            min_kp_anns=self.min_kp_anns if self.eval_annotation_filter else 0,
            category_ids=[1] if self.eval_annotation_filter else [],
        )
        return torch.utils.data.DataLoader(
            eval_data, batch_size=self.batch_size, shuffle=False,
            pin_memory=self.pin_memory, num_workers=self.loader_workers, drop_last=False,
            collate_fn=collate_images_anns_meta)

    def metrics(self):
        return [metric.Coco(
            COCO(self.eval_annotations),
            max_per_image=20,
            category_ids=[1],
            iou_type='keypoints',
            keypoint_oks_sigmas=ANIMAL_SIGMAS,
        )]



import argparse
import logging
import math
import torch
import torchvision.models
from . import effnetv2
from . import bottleneck_transformer

LOG = logging.getLogger(__name__)


class BaseNetwork(torch.nn.Module):
    """Common base network.

    :param name: a short name for the base network, e.g. resnet50
    :param stride: total stride from input to output
    :param out_features: number of output features
    """

    def __init__(self, name: str, *, stride: int, out_features: int):
        super().__init__()
        self.name = name
        self.stride = stride
        self.out_features = out_features
        LOG.info('%s: stride = %d, output features = %d', name, stride, out_features)

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        """Command line interface (CLI) to extend argument parser."""

    @classmethod
    def configure(cls, args: argparse.Namespace):
        """Take the parsed argument parser output and configure class variables."""

    def forward(self, x):
        raise NotImplementedError


class ShuffleNetV2(BaseNetwork):
    pretrained = True

    def __init__(self, name, torchvision_shufflenetv2, out_features=2048):
        super().__init__(name, stride=16, out_features=out_features)

        base_vision = torchvision_shufflenetv2(self.pretrained)
        self.conv1 = base_vision.conv1
        # base_vision.maxpool
        self.stage2 = base_vision.stage2
        self.stage3 = base_vision.stage3
        self.stage4 = base_vision.stage4
        self.conv5 = base_vision.conv5

    def forward(self, x):
        x = self.conv1(x)
        x = self.stage2(x)
        x = self.stage3(x)
        x = self.stage4(x)
        x = self.conv5(x)
        return x

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        group = parser.add_argument_group('ShuffleNetv2')
        assert cls.pretrained
        group.add_argument('--shufflenetv2-no-pretrain', dest='shufflenetv2_pretrained',
                           default=True, action='store_false',
                           help='use randomly initialized models')

    @classmethod
    def configure(cls, args: argparse.Namespace):
        cls.pretrained = args.shufflenetv2_pretrained


class Resnet(BaseNetwork):
    pretrained = True
    pool0_stride = 0
    input_conv_stride = 2
    input_conv2_stride = 0
    remove_last_block = False
    block5_dilation = 1

    def __init__(self, name, torchvision_resnet, out_features=2048):
        modules = list(torchvision_resnet(self.pretrained).children())
        stride = 32

        input_modules = modules[:4]

        # input pool
        if self.pool0_stride:
            if self.pool0_stride != 2:
                # pylint: disable=protected-access
                input_modules[3].stride = torch.nn.modules.utils._pair(self.pool0_stride)
                stride = int(stride * 2 / self.pool0_stride)
        else:
            input_modules.pop(3)
            stride //= 2

        # input conv
        if self.input_conv_stride != 2:
            # pylint: disable=protected-access
            input_modules[0].stride = torch.nn.modules.utils._pair(self.input_conv_stride)
            stride = int(stride * 2 / self.input_conv_stride)

        # optional use a conv in place of the max pool
        if self.input_conv2_stride:
            assert not self.pool0_stride  # this is only intended as a replacement for maxpool
            channels = input_modules[0].out_channels
            conv2 = torch.nn.Sequential(
                torch.nn.Conv2d(channels, channels, 3, 2, 1, bias=False),
                torch.nn.BatchNorm2d(channels),
                torch.nn.ReLU(inplace=True),
            )
            input_modules.append(conv2)
            stride *= 2
            LOG.debug('replaced max pool with [3x3 conv, bn, relu] with %d channels', channels)

        # block 5
        block5 = modules[7]
        if self.remove_last_block:
            block5 = None
            stride //= 2
            out_features //= 2

        if self.block5_dilation != 1:
            stride //= 2
            for m in block5.modules():
                if not isinstance(m, torch.nn.Conv2d):
                    continue

                # also must be changed for the skip-conv that has kernel=1
                m.stride = torch.nn.modules.utils._pair(1)

                if m.kernel_size[0] == 1:
                    continue

                m.dilation = torch.nn.modules.utils._pair(self.block5_dilation)
                padding = (m.kernel_size[0] - 1) // 2 * self.block5_dilation
                m.padding = torch.nn.modules.utils._pair(padding)

        super().__init__(name, stride=stride, out_features=out_features)
        self.input_block = torch.nn.Sequential(*input_modules)
        self.block2 = modules[4]
        self.block3 = modules[5]
        self.block4 = modules[6]
        self.block5 = block5

    def forward(self, x):
        x = self.input_block(x)
        x = self.block2(x)
        x = self.block3(x)
        x = self.block4(x)
        x = self.block5(x)
        return x

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        group = parser.add_argument_group('ResNet')
        assert cls.pretrained
        group.add_argument('--resnet-no-pretrain', dest='resnet_pretrained',
                           default=True, action='store_false',
                           help='use randomly initialized models')
        group.add_argument('--resnet-pool0-stride',
                           default=cls.pool0_stride, type=int,
                           help='stride of zero removes the pooling op')
        group.add_argument('--resnet-input-conv-stride',
                           default=cls.input_conv_stride, type=int,
                           help='stride of the input convolution')
        group.add_argument('--resnet-input-conv2-stride',
                           default=cls.input_conv2_stride, type=int,
                           help='stride of the optional 2nd input convolution')
        group.add_argument('--resnet-block5-dilation',
                           default=cls.block5_dilation, type=int,
                           help='use dilated convs in block5')
        assert not cls.remove_last_block
        group.add_argument('--resnet-remove-last-block',
                           default=False, action='store_true',
                           help='create a network without the last block')

    @classmethod
    def configure(cls, args: argparse.Namespace):
        cls.pretrained = args.resnet_pretrained
        cls.pool0_stride = args.resnet_pool0_stride
        cls.input_conv_stride = args.resnet_input_conv_stride
        cls.input_conv2_stride = args.resnet_input_conv2_stride
        cls.block5_dilation = args.resnet_block5_dilation
        cls.remove_last_block = args.resnet_remove_last_block


class InvertedResidualK(torch.nn.Module):
    """Based on torchvision.models.shufflenet.InvertedResidual."""

    def __init__(self, inp, oup, first_in_stage, *,
                 stride=1, layer_norm, non_linearity, dilation=1, kernel_size=3):
        super().__init__()
        assert (stride != 1 or dilation != 1 or inp != oup) or not first_in_stage
        LOG.debug('InvResK: %d %d %s, stride=%d, dilation=%d',
                  inp, oup, first_in_stage, stride, dilation)

        self.first_in_stage = first_in_stage
        branch_features = oup // 2
        padding = (kernel_size - 1) // 2 * dilation

        self.branch1 = None
        if self.first_in_stage:
            self.branch1 = torch.nn.Sequential(
                self.depthwise_conv(inp, inp,
                                    kernel_size=kernel_size, stride=stride,
                                    padding=padding, dilation=dilation),
                layer_norm(inp),
                torch.nn.Conv2d(inp, branch_features,
                                kernel_size=1, stride=1, padding=0, bias=False),
                layer_norm(branch_features),
                non_linearity(),
            )

        self.branch2 = torch.nn.Sequential(
            torch.nn.Conv2d(inp if first_in_stage else branch_features, branch_features,
                            kernel_size=1, stride=1, padding=0, bias=False),
            layer_norm(branch_features),
            non_linearity(),
            self.depthwise_conv(branch_features, branch_features,
                                kernel_size=kernel_size, stride=stride,
                                padding=padding, dilation=dilation),
            layer_norm(branch_features),
            torch.nn.Conv2d(branch_features, branch_features,
                            kernel_size=1, stride=1, padding=0, bias=False),
            layer_norm(branch_features),
            non_linearity(),
        )

    @staticmethod
    def depthwise_conv(in_f, out_f, kernel_size, stride=1, padding=0, bias=False, dilation=1):
        return torch.nn.Conv2d(in_f, out_f, kernel_size, stride, padding,
                               bias=bias, groups=in_f, dilation=dilation)

    def forward(self, x):
        if self.branch1 is None:
            x1, x2 = x.chunk(2, dim=1)
            out = torch.cat((x1, self.branch2(x2)), dim=1)
        else:
            out = torch.cat((self.branch1(x), self.branch2(x)), dim=1)

        out = torchvision.models.shufflenetv2.channel_shuffle(out, 2)

        return out


class ShuffleNetV2K(BaseNetwork):
    """Based on torchvision.models.ShuffleNetV2 where
    the kernel size in stages 2,3,4 is 5 instead of 3."""
    input_conv2_stride = 0
    input_conv2_outchannels = None
    layer_norm = None
    stage4_dilation = 1
    kernel_width = 5
    conv5_as_stage = False
    non_linearity = None

    def __init__(self, name, stages_repeats, stages_out_channels):
        layer_norm = ShuffleNetV2K.layer_norm
        if layer_norm is None:
            layer_norm = torch.nn.BatchNorm2d
        non_linearity = ShuffleNetV2K.non_linearity
        if non_linearity is None:
            non_linearity = lambda: torch.nn.ReLU(inplace=True)

        if len(stages_repeats) != 3:
            raise ValueError('expected stages_repeats as list of 3 positive ints')
        if len(stages_out_channels) != 5:
            raise ValueError('expected stages_out_channels as list of 5 positive ints')
        _stage_out_channels = stages_out_channels

        stride = 16  # in the default configuration
        input_modules = []
        input_channels = 3
        output_channels = _stage_out_channels[0]
        conv1 = torch.nn.Sequential(
            torch.nn.Conv2d(input_channels, output_channels, 3, 2, 1, bias=False),
            layer_norm(output_channels),
            non_linearity(),
        )
        input_modules.append(conv1)
        input_channels = output_channels

        # optional use a conv in place of the max pool
        if self.input_conv2_stride:
            output_channels = self.input_conv2_outchannels or input_channels
            conv2 = torch.nn.Sequential(
                torch.nn.Conv2d(input_channels, output_channels, 3, 2, 1, bias=False),
                layer_norm(output_channels),
                non_linearity(),
            )
            input_modules.append(conv2)
            stride *= 2
            input_channels = output_channels
            LOG.debug('replaced max pool with [3x3 conv, bn, relu]')

        stages = []
        for repeats, output_channels, dilation in zip(
                stages_repeats, _stage_out_channels[1:], [1, 1, self.stage4_dilation]):
            stage_stride = 2 if dilation == 1 else 1
            stride = int(stride * stage_stride / 2)
            seq = [InvertedResidualK(input_channels, output_channels, True,
                                     kernel_size=self.kernel_width,
                                     layer_norm=layer_norm,
                                     non_linearity=non_linearity,
                                     dilation=dilation,
                                     stride=stage_stride)]
            for _ in range(repeats - 1):
                seq.append(InvertedResidualK(output_channels, output_channels, False,
                                             kernel_size=self.kernel_width,
                                             layer_norm=layer_norm,
                                             non_linearity=non_linearity,
                                             dilation=dilation))
            stages.append(torch.nn.Sequential(*seq))
            input_channels = output_channels

        output_channels = _stage_out_channels[-1]
        if self.conv5_as_stage:
            # Two stages are about the same number of parameters as one
            # convolution.
            # Conv: 1392*1392
            # Two Stages: 4 * 696*696 + 2 * 5^2*696
            use_first_in_stage = input_channels != output_channels
            conv5 = torch.nn.Sequential(
                InvertedResidualK(input_channels, output_channels, use_first_in_stage,
                                  kernel_size=self.kernel_width,
                                  layer_norm=layer_norm,
                                  non_linearity=non_linearity,
                                  dilation=self.stage4_dilation),
                InvertedResidualK(output_channels, output_channels, False,
                                  kernel_size=self.kernel_width,
                                  layer_norm=layer_norm,
                                  non_linearity=non_linearity,
                                  dilation=self.stage4_dilation),
            )
        else:
            conv5 = torch.nn.Sequential(
                torch.nn.Conv2d(input_channels, output_channels, 1, 1, 0, bias=False),
                layer_norm(output_channels),
                non_linearity(),
            )

        super().__init__(name, stride=stride, out_features=output_channels)
        self.input_block = torch.nn.Sequential(*input_modules)
        self.stage2 = stages[0]
        self.stage3 = stages[1]
        self.stage4 = stages[2]
        self.conv5 = conv5

    def forward(self, x):
        x = self.input_block(x)
        x = self.stage2(x)
        x = self.stage3(x)
        x = self.stage4(x)
        x = self.conv5(x)
        return x

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        group = parser.add_argument_group('shufflenetv2k')
        group.add_argument('--shufflenetv2k-input-conv2-stride',
                           default=cls.input_conv2_stride, type=int,
                           help='stride of the optional 2nd input convolution')
        group.add_argument('--shufflenetv2k-input-conv2-outchannels',
                           default=cls.input_conv2_outchannels, type=int,
                           help='out channels of the optional 2nd input convolution')
        group.add_argument('--shufflenetv2k-stage4-dilation',
                           default=cls.stage4_dilation, type=int,
                           help='dilation factor of stage 4')
        group.add_argument('--shufflenetv2k-kernel',
                           default=cls.kernel_width, type=int,
                           help='kernel width')
        assert not cls.conv5_as_stage
        group.add_argument('--shufflenetv2k-conv5-as-stage',
                           default=False, action='store_true')

        layer_norm_group = group.add_mutually_exclusive_group()
        layer_norm_group.add_argument('--shufflenetv2k-instance-norm',
                                      default=False, action='store_true')
        layer_norm_group.add_argument('--shufflenetv2k-group-norm',
                                      default=False, action='store_true')

        non_linearity_group = group.add_mutually_exclusive_group()
        non_linearity_group.add_argument('--shufflenetv2k-leaky-relu',
                                         default=False, action='store_true')

    @classmethod
    def configure(cls, args: argparse.Namespace):
        cls.input_conv2_stride = args.shufflenetv2k_input_conv2_stride
        cls.input_conv2_outchannels = args.shufflenetv2k_input_conv2_outchannels
        cls.stage4_dilation = args.shufflenetv2k_stage4_dilation
        cls.kernel_width = args.shufflenetv2k_kernel
        cls.conv5_as_stage = args.shufflenetv2k_conv5_as_stage

        # layer norms
        if args.shufflenetv2k_instance_norm:
            cls.layer_norm = lambda x: torch.nn.InstanceNorm2d(
                x, affine=True, track_running_stats=True)
        if args.shufflenetv2k_group_norm:
            cls.layer_norm = lambda x: torch.nn.GroupNorm(
                (32 if x % 32 == 0 else 29) if x > 100 else 4, x)

        # non-linearities
        if args.shufflenetv2k_leaky_relu:
            cls.non_linearity = lambda: torch.nn.LeakyReLU(inplace=True)


class MobileNetV2(BaseNetwork):
    pretrained = True

    def __init__(self, name, torchvision_mobilenetv2, out_features=1280):
        super().__init__(name, stride=32, out_features=out_features)
        base_vision = torchvision_mobilenetv2(self.pretrained)
        self.backbone = list(base_vision.children())[0]  # remove output classifier

    def forward(self, x):
        x = self.backbone(x)
        return x

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        group = parser.add_argument_group('MobileNetV2')
        assert cls.pretrained
        group.add_argument('--mobilenetv2-no-pretrain', dest='mobilenetv2_pretrained',
                           default=True, action='store_false',
                           help='use randomly initialized models')

    @classmethod
    def configure(cls, args: argparse.Namespace):
        cls.pretrained = args.mobilenetv2_pretrained


class MobileNetV3(BaseNetwork):
    pretrained = True

    def __init__(self, name, torchvision_mobilenetv3, out_features=960):
        super().__init__(name, stride=16, out_features=out_features)
        base_vision = torchvision_mobilenetv3(self.pretrained)

        self.backbone = list(base_vision.children())[0]  # remove output classifier
        # remove stride from input block
        input_conv = list(self.backbone)[0][0]
        input_conv.stride = torch.nn.modules.utils._pair(1)  # pylint: disable=protected-access

    def forward(self, x):
        x = self.backbone(x)
        return x

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        group = parser.add_argument_group('MobileNetV3')
        assert cls.pretrained
        group.add_argument('--mobilenetv3-no-pretrain', dest='mobilenetv3_pretrained',
                           default=True, action='store_false',
                           help='use randomly initialized models')

    @classmethod
    def configure(cls, args: argparse.Namespace):
        cls.pretrained = args.mobilenetv3_pretrained


class SqueezeNet(BaseNetwork):
    pretrained = True

    def __init__(self, name, torchvision_squeezenet, out_features=512):
        super().__init__(name, stride=16, out_features=out_features)
        base_vision = torchvision_squeezenet(self.pretrained)

        for m in base_vision.modules():
            # adjust padding on all maxpool layers
            if isinstance(m, (torch.nn.MaxPool2d,)) and m.padding != 1:
                LOG.debug('adjusting maxpool2d padding to 1 from padding=%d, kernel=%d, stride=%d',
                          m.padding, m.kernel_size, m.stride)
                m.padding = 1

            # adjust padding on some conv2d (only the first one)
            if isinstance(m, (torch.nn.Conv2d,)):
                target_padding = (m.kernel_size[0] - 1) // 2
                if m.padding[0] != target_padding:
                    LOG.debug('adjusting conv2d padding to %d (kernel=%d, padding=%d)',
                              target_padding, m.kernel_size, m.padding)
                    m.padding = torch.nn.modules.utils._pair(target_padding)  # pylint: disable=protected-access

        self.backbone = list(base_vision.children())[0]  # remove output classifier

    def forward(self, x):
        x = self.backbone(x)
        return x

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        group = parser.add_argument_group('SqueezeNet')
        assert cls.pretrained
        group.add_argument('--squeezenet-no-pretrain', dest='squeezenet_pretrained',
                           default=True, action='store_false',
                           help='use randomly initialized models')

    @classmethod
    def configure(cls, args: argparse.Namespace):
        cls.pretrained = args.squeezenet_pretrained


class FPN(torch.nn.Module):
    """ Feature Pyramid Network (https://arxiv.org/abs/1612.03144), modified to only
    refine and return the feature map of a single pyramid level.

    This implementation is more computationally efficient than torchvision's
    FeaturePyramidNetwork when only a single feature map is needed, as it avoids refining
    (i.e. applying a 3x3 conv on) feature maps that aren't used later on.

    For example, for Swin, if only the feature map of stride 8 (fpn_level=2) is needed,
    the feature maps of stride 4, 16 and 32 won't get refined with this implementation.
    """

    def __init__(self, in_channels, out_channels, fpn_level=3):

        super().__init__()

        self.num_upsample_ops = len(in_channels) - fpn_level

        self.lateral_convs = torch.nn.ModuleList()

        # Start from the higher-level features (start from the smaller feature maps)
        for i in range(1, 2 + self.num_upsample_ops):
            lateral_conv = torch.nn.Conv2d(in_channels[-i], out_channels, 1)
            self.lateral_convs.append(lateral_conv)

        # Only one single refine conv for the largest feature map
        self.refine_conv = torch.nn.Conv2d(out_channels, out_channels, 3, padding=1)

    def forward(self, inputs):
        # FPN top-down pathway
        # Start from the higher-level features (start from the smaller feature maps)
        laterals = [lateral_conv(x)
                    for lateral_conv, x in zip(self.lateral_convs, inputs[::-1])]

        for i in range(1, 1 + self.num_upsample_ops):
            laterals[i] += torch.nn.functional.interpolate(
                laterals[i - 1], size=laterals[i].shape[2:], mode='nearest')

        x = self.refine_conv(laterals[-1])
        return x


class SwinTransformer(BaseNetwork):
    """Swin Transformer, with optional FPN and input upsampling to obtain higher resolution
    feature maps"""
    pretrained = True
    drop_path_rate = 0.2
    input_upsample = False
    use_fpn = False
    fpn_level = 3
    fpn_out_channels = None

    def __init__(self, name, swin_net):
        embed_dim = swin_net().embed_dim

        if not self.use_fpn or self.fpn_out_channels is None:
            self.out_features = 8 * embed_dim
        else:
            self.out_features = self.fpn_out_channels

        stride = 32

        if self.input_upsample:
            LOG.debug('swin input upsampling')
            stride //= 2

        if self.use_fpn:
            LOG.debug('swin output FPN level: %d', self.fpn_level)
            stride //= 2 ** (4 - self.fpn_level)

        super().__init__(name, stride=stride, out_features=self.out_features)

        self.input_upsample_op = None
        if self.input_upsample:
            self.input_upsample_op = torch.nn.Upsample(scale_factor=2)

        if not self.use_fpn:
            out_indices = [3, ]
        else:
            out_indices = list(range(self.fpn_level - 1, 4))

        self.backbone = swin_net(pretrained=self.pretrained,
                                 drop_path_rate=self.drop_path_rate,
                                 out_indices=out_indices)

        self.fpn = None
        if self.use_fpn:
            self.fpn = FPN([embed_dim, 2 * embed_dim, 4 * embed_dim, 8 * embed_dim],
                           self.out_features, self.fpn_level)

    def forward(self, x):
        if self.input_upsample_op is not None:
            x = self.input_upsample_op(x)

        outs = self.backbone(x)

        if self.fpn is not None:
            x = self.fpn(outs)
        else:
            x = outs[-1]

        return x

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        group = parser.add_argument_group('SwinTransformer')

        group.add_argument('--swin-drop-path-rate', default=cls.drop_path_rate, type=float,
                           help="drop path (stochastic depth) rate")

        group.add_argument('--swin-input-upsample', default=False, action='store_true',
                           help='scales input image by a factor of 2 for higher res feature maps')

        group.add_argument('--swin-use-fpn', default=False, action='store_true',
                           help='adds a FPN after the Swin network '
                                'to obtain higher res feature maps')

        group.add_argument('--swin-fpn-out-channels',
                           default=cls.fpn_out_channels, type=int,
                           help='output channels of the FPN (None to use the '
                                'default number of channels of the Swin network)')

        group.add_argument('--swin-fpn-level',
                           default=cls.fpn_level, type=int,
                           help='FPN pyramid level, must be between 1 '
                                '(highest resolution) and 4 (lowest resolution)')

        group.add_argument('--swin-no-pretrain', dest='swin_pretrained',
                           default=True, action='store_false',
                           help='use randomly initialized models')

    @classmethod
    def configure(cls, args: argparse.Namespace):
        cls.drop_path_rate = args.swin_drop_path_rate
        cls.input_upsample = args.swin_input_upsample
        cls.use_fpn = args.swin_use_fpn
        cls.fpn_out_channels = args.swin_fpn_out_channels
        cls.fpn_level = args.swin_fpn_level
        cls.pretrained = args.swin_pretrained


class XCiT(BaseNetwork):
    pretrained = True
    out_channels = None
    out_maxpool = False

    def __init__(self, name, xcit_net):
        embed_dim = xcit_net().embed_dim
        patch_size = xcit_net().patch_size
        has_projection = isinstance(self.out_channels, int)
        self.out_channels = self.out_channels if has_projection else embed_dim

        stride = patch_size * 2 if self.out_maxpool else patch_size

        super().__init__(name, stride=stride, out_features=self.out_channels)

        self.backbone = xcit_net(pretrained=self.pretrained)

        if has_projection:
            LOG.debug('adding output projection to %d channels', self.out_features)
            self.out_projection = torch.nn.Conv2d(
                embed_dim, self.out_features, kernel_size=1, stride=1)
        else:
            LOG.debug('no output projection')
            self.out_projection = torch.nn.Identity()

        if self.out_maxpool:
            LOG.debug('output max-pooling')
            self.out_block = torch.nn.Sequential(
                self.out_projection,
                torch.nn.MaxPool2d(kernel_size=2, stride=2, ceil_mode=True)
            )
        else:
            self.out_block = torch.nn.Sequential(self.out_projection)

    def forward(self, x):
        x = self.backbone(x)
        x = self.out_block(x)
        return x

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        group = parser.add_argument_group('XCiT')
        group.add_argument('--xcit-out-channels',
                           default=cls.out_channels, type=int,
                           help='number of output channels for optional projection layer '
                                '(None for no projection layer)')

        group.add_argument('--xcit-out-maxpool', default=False, action='store_true',
                           help='adds max-pooling to backbone output feature map')

        group.add_argument('--xcit-no-pretrain', dest='xcit_pretrained',
                           default=True, action='store_false',
                           help='use randomly initialized models')

    @classmethod
    def configure(cls, args: argparse.Namespace):
        cls.out_channels = args.xcit_out_channels
        cls.out_maxpool = args.xcit_out_maxpool
        cls.pretrained = args.xcit_pretrained


class EffNetV2(BaseNetwork):
    def __init__(self, name, configuration, stride):
        backbone = effnetv2.EffNetV2(configuration)
        super().__init__(name, stride=stride, out_features=backbone.output_channel)
        self.backbone = backbone
        self.backbone._initialize_weights()

    def forward(self, x):
        x = self.backbone.forward(x)
        return x

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        pass

    @classmethod
    def configure(cls, args: argparse.Namespace):
        pass


class BotNet(BaseNetwork):
    input_image_size = 640

    def __init__(self, name, out_features=2048):
        super().__init__(name, stride=8, out_features=out_features)

        layer = bottleneck_transformer.BottleStack(
            dim=256,
            fmap_size=int(math.ceil(self.input_image_size / 4)),  # default img size is 640 x 640
            dim_out=2048,
            proj_factor=4,
            downsample=True,
            heads=4,
            dim_head=128,
            rel_pos_emb=True,
            activation=torch.nn.ReLU()
        )

        resnet = torchvision.models.resnet50()

        # model surgery
        resnet_parts = list(resnet.children())
        self.backbone = torch.nn.Sequential(
            *resnet_parts[:5],
            layer,
        )

    def forward(self, x):
        x = self.backbone.forward(x)
        return x

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        group = parser.add_argument_group('BotNet')
        group.add_argument('--botnet-input-image-size',
                           default=cls.input_image_size, type=int,
                           help='Input image size. Needs to be the same for training and'
                           ' prediction, as BotNet only accepts fixed input sizes')

    @classmethod
    def configure(cls, args: argparse.Namespace):
        cls.input_image_size = args.botnet_input_image_size



import argparse
from typing import List

import numpy as np

from .. import annotation
from .decoder import Decoder
from .track_annotation import TrackAnnotation
from ..signal import Signal


class TrackBase(Decoder):
    single_pose_threshold = 0.3
    multi_pose_threshold = 0.2
    multi_pose_n = 3
    minimum_threshold = 0.1
    simplify_good_ids = True
    track_visualizer = None

    def __init__(self):
        super().__init__()

        self.active: List[TrackAnnotation] = []
        self.frame_number = 0

        self.simplified_track_id_map = {}
        self.simplified_last_track_id = 0

        Signal.subscribe('eval_reset', self.reset)

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        """Command line interface (CLI) to extend argument parser."""
        group = parser.add_argument_group('Decoder for tracking')
        group.add_argument('--tr-single-pose-threshold',
                           default=cls.single_pose_threshold, type=float,
                           help='Single-pose threshold for tracking.')
        group.add_argument('--tr-multi-pose-threshold',
                           default=cls.multi_pose_threshold, type=float,
                           help='multi-pose threshold for tracking.')
        group.add_argument('--tr-multi-pose-n',
                           default=cls.multi_pose_n, type=float,
                           help='multi-pose n for tracking.')
        group.add_argument('--tr-minimum-threshold',
                           default=cls.minimum_threshold, type=float,
                           help='minimum-pose threshold for tracking.')

    @classmethod
    def configure(cls, args: argparse.Namespace):
        """Take the parsed argument parser output and configure class variables."""
        cls.single_pose_threshold = args.tr_single_pose_threshold
        cls.multi_pose_threshold = args.tr_multi_pose_threshold
        cls.multi_pose_n = args.tr_multi_pose_n
        cls.minimum_threshold = args.tr_minimum_threshold

    @classmethod
    def factory(cls, head_metas) -> List['Decoder']:
        """Create instances of an implementation."""
        raise NotImplementedError

    def __call__(self, fields, *, initial_annotations=None) -> List[annotation.Base]:
        """For single image, from fields to annotations."""
        raise NotImplementedError

    def simplify_ids(self, ids):
        out = []
        for id_ in ids:
            if id_ not in self.simplified_track_id_map:
                self.simplified_last_track_id += 1
                self.simplified_track_id_map[id_] = self.simplified_last_track_id

            out.append(self.simplified_track_id_map[id_])
        return out

    def reset(self):
        self.active = []
        self.frame_number = 0

        self.simplified_track_id_map = {}
        self.simplified_last_track_id = 0

    def prune_active(self, frame_number):
        self.active = [t for t in self.active
                       if frame_number - t.frame_pose[-1][0] <= 33]
        self.active = [t for t in self.active
                       if frame_number - t.frame_pose[-1][0] == 1 or len(t.frame_pose) > 2]

    def annotations(self, frame_number):
        tracks = [t for t in self.active if t.frame_pose[-1][0] == frame_number]
        tracks = [t for t in tracks if self.track_is_good(t, frame_number)]
        if not tracks:
            return []

        ids = [t.id_ for t in tracks]
        if self.simplify_good_ids:
            ids = self.simplify_ids(ids)
        annotations = [t.frame_pose[-1][1] for t in tracks]
        for ann, id_ in zip(annotations, ids):
            ann.id_ = id_
        return annotations

    def tag_ignore_region(self, frame_number, gt_anns):
        pose_annotations = [track.frame_pose[-1][1]
                            for track in self.active
                            if track.frame_pose[-1][0] == frame_number]
        crowd_annotations = [a for a in gt_anns if a['iscrowd']]

        def point_in_polygon(x, y, poly_x, poly_y):
            inside = False
            for x1, x2, y1, y2 in zip(poly_x[:-1], poly_x[1:], poly_y[:-1], poly_y[1:]):
                if min(y1, y2) > y or max(y1, y2) < y:
                    continue
                lx = x1 + (x2 - x1) * (y - y1) / (y2 - y1)
                if lx < x:
                    inside = not inside
            return inside

        def pa_in_ca(pose_annotation, crowd_annotation):
            pose = pose_annotation.data
            poly = crowd_annotation['keypoints'][:, :2].tolist()
            # close polygon
            poly.append(poly[0])
            xs = [x for x, _ in poly]
            ys = [y for _, y in poly]

            kp_order = np.argsort(pose[:, 2])[::-1]
            if all(point_in_polygon(kp[0], kp[1], xs, ys)
                   for kp in pose[kp_order[:3]] if kp[2] > 0.05):
                return True

            return False

        for pa in pose_annotations:
            pa.ignore_region = any(pa_in_ca(pa, ca) for ca in crowd_annotations)

    def track_is_viable(self, track, frame_number):
        if frame_number > track.frame_pose[-1][0] + 33:
            return False

        if any(track.pose_score(frame_number - i) > self.multi_pose_threshold for i in range(33)):
            return True

        return False

    def track_is_good(self, track, frame_number):
        for i in range(4):
            pose = track.pose(frame_number - i)
            if pose is None:
                continue
            if getattr(pose, 'ignore_region', False):
                return False

        if not self.track_is_viable(track, frame_number):
            return False

        if all(track.pose_score(frame_number - i) < self.single_pose_threshold
               for i in range(6)) and \
           sum(1
               for i in range(6)
               if track.pose_score(frame_number - i) > self.multi_pose_threshold
               ) < self.multi_pose_n:
            return False

        assert self.minimum_threshold >= 0.0  # make sure to return False when pose is None
        if track.pose_score(frame_number) <= self.minimum_threshold:
            return False

        return True



"""Train a pifpaf network."""

import argparse
import copy
import datetime
import logging
import os
import socket

import torch

from . import datasets, encoder, logger, network, optimize, plugin, show, visualizer
from . import __version__

LOG = logging.getLogger(__name__)


def default_output_file(args):
    base_name = args.basenet
    if not base_name:
        base_name, _, __ = os.path.basename(args.checkpoint).partition('-')

    now = datetime.datetime.now().strftime('%y%m%d-%H%M%S')
    out = 'outputs/{}-{}-{}'.format(base_name, now, args.dataset)

    # Slurm jobs might be stuck in the job queue and then started at exactly the
    # same time. Therefore we disambiguate with the Slurm job id.
    if os.getenv('SLURM_JOB_ID'):
        out += f'-slurm{os.getenv("SLURM_JOB_ID")}'

    if args.cocokp_square_edge != 385:
        out += '-edge{}'.format(args.cocokp_square_edge)
    if args.regression_loss != 'laplace':
        out += '-{}'.format(args.regression_loss)
    if args.r_smooth != 0.0:
        out += '-rsmooth{}'.format(args.r_smooth)
    if args.cocokp_orientation_invariant or args.cocokp_extended_scale:
        out += '-'
        if args.cocokp_orientation_invariant:
            out += 'o{:02.0f}'.format(args.cocokp_orientation_invariant * 100.0)
        if args.cocokp_extended_scale:
            out += 's'

    return out + '.pkl'


def cli():
    parser = argparse.ArgumentParser(
        prog='python3 -m openpifpaf.train',
        usage='%(prog)s [options]',
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument('--version', action='version',
                        version='OpenPifPaf {version}'.format(version=__version__))
    parser.add_argument('-o', '--output', default=None,
                        help='output file')
    parser.add_argument('--disable-cuda', action='store_true',
                        help='disable CUDA')
    parser.add_argument('--ddp', default=False, action='store_true',
                        help='[experimental] DistributedDataParallel')
    parser.add_argument('--local_rank', default=None, type=int,
                        help='[experimental] for torch.distributed.launch')
    parser.add_argument('--no-sync-batchnorm', dest='sync_batchnorm',
                        default=True, action='store_false',
                        help='[experimental] in ddp, to not use syncbatchnorm')

    logger.cli(parser)
    network.Factory.cli(parser)
    network.losses.Factory.cli(parser)
    network.Trainer.cli(parser)
    encoder.cli(parser)
    optimize.cli(parser)
    datasets.cli(parser)
    show.cli(parser)
    visualizer.cli(parser)

    args = parser.parse_args()

    logger.configure(args, LOG)
    if args.log_stats:
        logging.getLogger('openpifpaf.stats').setLevel(logging.DEBUG)

    # DDP with SLURM
    slurm_process_id = os.environ.get('SLURM_PROCID')
    if args.ddp and slurm_process_id is not None:
        if torch.cuda.device_count() > 1:
            LOG.warning('Expected one GPU per SLURM task but found %d. '
                        'Try with "srun --gpu-bind=closest ...". Still trying.',
                        torch.cuda.device_count())

        # if there is more than one GPU available, assume that other SLURM tasks
        # have access to the same GPUs and assign GPUs uniquely by slurm_process_id
        args.local_rank = (int(slurm_process_id) % torch.cuda.device_count()
                           if torch.cuda.device_count() > 0 else 0)

        os.environ['RANK'] = slurm_process_id
        if not os.environ.get('WORLD_SIZE') and os.environ.get('SLURM_NTASKS'):
            os.environ['WORLD_SIZE'] = os.environ.get('SLURM_NTASKS')

        LOG.info('found SLURM process id: %s', slurm_process_id)
        LOG.info('distributed env: master=%s port=%s rank=%s world=%s, '
                 'local rank (GPU)=%d',
                 os.environ.get('MASTER_ADDR'), os.environ.get('MASTER_PORT'),
                 os.environ.get('RANK'), os.environ.get('WORLD_SIZE'),
                 args.local_rank)

    # add args.device
    args.device = torch.device('cpu')
    args.pin_memory = False
    if not args.disable_cuda and torch.cuda.is_available():
        args.device = torch.device('cuda')
        args.pin_memory = True
    LOG.info('neural network device: %s (CUDA available: %s, count: %d)',
             args.device, torch.cuda.is_available(), torch.cuda.device_count())

    # output
    if args.output is None:
        args.output = default_output_file(args)
        os.makedirs('outputs', exist_ok=True)

    network.Factory.configure(args)
    network.losses.Factory.configure(args)
    network.Trainer.configure(args)
    encoder.configure(args)
    datasets.configure(args)
    show.configure(args)
    visualizer.configure(args)

    return args


def main():
    args = cli()

    datamodule = datasets.factory(args.dataset)

    net_cpu, start_epoch = network.Factory().factory(head_metas=datamodule.head_metas)
    loss = network.losses.Factory().factory(datamodule.head_metas)

    checkpoint_shell = None
    if not args.disable_cuda and torch.cuda.device_count() > 1 and not args.ddp:
        LOG.info('Multiple GPUs with DataParallel: %d', torch.cuda.device_count())
        checkpoint_shell = copy.deepcopy(net_cpu)
        net = torch.nn.DataParallel(net_cpu.to(device=args.device))
        loss = loss.to(device=args.device)
    elif not args.disable_cuda and torch.cuda.device_count() == 1 and not args.ddp:
        LOG.info('Single GPU training')
        checkpoint_shell = copy.deepcopy(net_cpu)
        net = net_cpu.to(device=args.device)
        loss = loss.to(device=args.device)
    elif not args.disable_cuda and torch.cuda.device_count() > 0:
        LOG.info('Multiple GPUs with DistributedDataParallel')
        assert not list(loss.parameters())
        assert torch.cuda.device_count() > 0
        checkpoint_shell = copy.deepcopy(net_cpu)
        torch.cuda.set_device(args.local_rank)
        torch.distributed.init_process_group(backend='nccl', init_method='env://')
        LOG.info('DDP: rank %d, world %d',
                 torch.distributed.get_rank(), torch.distributed.get_world_size())
        if args.sync_batchnorm:
            LOG.info('convert all batchnorms to syncbatchnorms')
            net_cpu = torch.nn.SyncBatchNorm.convert_sync_batchnorm(net_cpu)
        else:
            LOG.info('not converting batchnorms to syncbatchnorms')
        net = torch.nn.parallel.DistributedDataParallel(
            net_cpu.to(device=args.device),
            device_ids=[args.local_rank], output_device=args.local_rank,
            find_unused_parameters=isinstance(datamodule, datasets.MultiDataModule),
        )
        loss = loss.to(device=args.device)
    else:
        net = net_cpu

    logger.train_configure(args)
    train_loader = datamodule.train_loader()
    val_loader = datamodule.val_loader()
    if torch.distributed.is_initialized():
        train_loader = datamodule.distributed_sampler(train_loader)
        val_loader = datamodule.distributed_sampler(val_loader)

    optimizer = optimize.factory_optimizer(
        args, list(net.parameters()) + list(loss.parameters()))
    lr_scheduler = optimize.factory_lrscheduler(
        args, optimizer, len(train_loader), last_epoch=start_epoch)
    trainer = network.Trainer(
        net, loss, optimizer, args.output,
        checkpoint_shell=checkpoint_shell,
        lr_scheduler=lr_scheduler,
        device=args.device,
        model_meta_data={
            'args': vars(args),
            'version': __version__,
            'plugin_versions': plugin.versions(),
            'hostname': socket.gethostname(),
        },
    )
    trainer.loop(train_loader, val_loader, start_epoch=start_epoch)


if __name__ == '__main__':
    main()



from ..preprocess import Preprocess


class Encoders(Preprocess):
    def __init__(self, encoders):
        self.encoders = encoders

    def __call__(self, images, all_anns, metas):
        targets = [enc(images, all_anns, metas) for enc in self.encoders]
        meta = metas[0]
        meta['head_indices'] = [enc.meta.head_index for enc in self.encoders]
        return images, targets, meta
